# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=too-many-arguments
"""
This submodule contains the discrete-variable quantum operations that are the
core parameterized gates.
"""
# pylint:disable=abstract-method,arguments-differ,protected-access,invalid-overridden-method
import functools
import numpy as np

import pennylane as qml
from pennylane.operation import Operation
from .non_parametric_ops import Hadamard, PauliX, PauliY, PauliZ

stack_last = functools.partial(qml.math.stack, axis=-1)


def _can_replace(x, y):
    """
    Convenience function that returns true if x is close to y and if
    x does not require grad
    """
    return not qml.math.is_abstract(x) and not qml.math.requires_grad(x) and qml.math.allclose(x, y)


class RX(Operation):
    r"""
    The single qubit X rotation

    .. math:: R_x(\phi) = e^{-i\phi\sigma_x/2} = \begin{bmatrix}
                \cos(\phi/2) & -i\sin(\phi/2) \\
                -i\sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Number of dimensions per parameter: (0,)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_x(\phi)) = \frac{1}{2}\left[f(R_x(\phi+\pi/2)) - f(R_x(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_x(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    basis = "X"
    grad_method = "A"
    parameter_frequencies = [(1,)]

    def generator(self):
        return -0.5 * PauliX(wires=self.wires)

    def __init__(self, phi, wires, do_queue=True, id=None):
        super().__init__(phi, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(theta):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.RX.matrix`

        Args:
            theta (tensor_like or float): rotation angle

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.RX.compute_matrix(torch.tensor(0.5))
        tensor([[0.9689+0.0000j, 0.0000-0.2474j],
                [0.0000-0.2474j, 0.9689+0.0000j]])
        """
        c = qml.math.cos(theta / 2)
        s = qml.math.sin(theta / 2)

        if qml.math.get_interface(theta) == "tensorflow":
            c = qml.math.cast_like(c, 1j)
            s = qml.math.cast_like(s, 1j)

        # The following avoids casting an imaginary quantity to reals when backpropagating
        c = (1 + 0j) * c
        js = -1j * s
        return qml.math.stack([stack_last([c, js]), stack_last([js, c])], axis=-2)

    def adjoint(self):
        return RX(-self.data[0], wires=self.wires)

    def pow(self, z):
        return [RX(self.data[0] * z, wires=self.wires)]

    def _controlled(self, wire):
        return qml.CRX(*self.parameters, wires=wire + self.wires)

    def simplify(self):
        theta = self.data[0] % (4 * np.pi)

        if _can_replace(theta, 0):
            return qml.Identity(wires=self.wires)

        return RX(theta, wires=self.wires)

    def single_qubit_rot_angles(self):
        # RX(\theta) = RZ(-\pi/2) RY(\theta) RZ(\pi/2)
        pi_half = qml.math.ones_like(self.data[0]) * (np.pi / 2)
        return [pi_half, self.data[0], -pi_half]


class RY(Operation):
    r"""
    The single qubit Y rotation

    .. math:: R_y(\phi) = e^{-i\phi\sigma_y/2} = \begin{bmatrix}
                \cos(\phi/2) & -\sin(\phi/2) \\
                \sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Number of dimensions per parameter: (0,)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_y(\phi)) = \frac{1}{2}\left[f(R_y(\phi+\pi/2)) - f(R_y(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_y(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    basis = "Y"
    grad_method = "A"
    parameter_frequencies = [(1,)]

    def generator(self):
        return -0.5 * PauliY(wires=self.wires)

    def __init__(self, phi, wires, do_queue=True, id=None):
        super().__init__(phi, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(theta):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.RY.matrix`


        Args:
            theta (tensor_like or float): rotation angle

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.RY.compute_matrix(torch.tensor(0.5))
        tensor([[ 0.9689, -0.2474],
                [ 0.2474,  0.9689]])
        """

        c = qml.math.cos(theta / 2)
        s = qml.math.sin(theta / 2)
        if qml.math.get_interface(theta) == "tensorflow":
            c = qml.math.cast_like(c, 1j)
            s = qml.math.cast_like(s, 1j)
        # The following avoids casting an imaginary quantity to reals when backpropagating
        c = (1 + 0j) * c
        s = (1 + 0j) * s
        return qml.math.stack([stack_last([c, -s]), stack_last([s, c])], axis=-2)

    def adjoint(self):
        return RY(-self.data[0], wires=self.wires)

    def pow(self, z):
        return [RY(self.data[0] * z, wires=self.wires)]

    def _controlled(self, wire):
        return qml.CRY(*self.parameters, wires=wire + self.wires)

    def simplify(self):
        theta = self.data[0] % (4 * np.pi)

        if _can_replace(theta, 0):
            return qml.Identity(wires=self.wires)

        return RY(theta, wires=self.wires)

    def single_qubit_rot_angles(self):
        # RY(\theta) = RZ(0) RY(\theta) RZ(0)
        return [0.0, self.data[0], 0.0]


class RZ(Operation):
    r"""
    The single qubit Z rotation

    .. math:: R_z(\phi) = e^{-i\phi\sigma_z/2} = \begin{bmatrix}
                e^{-i\phi/2} & 0 \\
                0 & e^{i\phi/2}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Number of dimensions per parameter: (0,)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_z(\phi)) = \frac{1}{2}\left[f(R_z(\phi+\pi/2)) - f(R_z(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_z(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    basis = "Z"
    grad_method = "A"
    parameter_frequencies = [(1,)]

    def generator(self):
        return -0.5 * PauliZ(wires=self.wires)

    def __init__(self, phi, wires, do_queue=True, id=None):
        super().__init__(phi, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(theta):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.RZ.matrix`

        Args:
            theta (tensor_like or float): rotation angle

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.RZ.compute_matrix(torch.tensor(0.5))
        tensor([[0.9689-0.2474j, 0.0000+0.0000j],
                [0.0000+0.0000j, 0.9689+0.2474j]])
        """
        if qml.math.get_interface(theta) == "tensorflow":
            p = qml.math.exp(-0.5j * qml.math.cast_like(theta, 1j))
            z = qml.math.zeros_like(p)

            return qml.math.stack([stack_last([p, z]), stack_last([z, qml.math.conj(p)])], axis=-2)

        signs = qml.math.array([-1, 1], like=theta)
        arg = 0.5j * theta

        if qml.math.ndim(arg) == 0:
            return qml.math.diag(qml.math.exp(arg * signs))

        diags = qml.math.exp(qml.math.outer(arg, signs))
        return diags[:, :, np.newaxis] * qml.math.cast_like(qml.math.eye(2, like=diags), diags)

    @staticmethod
    def compute_eigvals(theta):  # pylint: disable=arguments-differ
        r"""Eigenvalues of the operator in the computational basis (static method).

        If :attr:`diagonalizing_gates` are specified and implement a unitary :math:`U^{\dagger}`,
        the operator can be reconstructed as

        .. math:: O = U \Sigma U^{\dagger},

        where :math:`\Sigma` is the diagonal matrix containing the eigenvalues.

        Otherwise, no particular order for the eigenvalues is guaranteed.

        .. seealso:: :meth:`~.RZ.eigvals`


        Args:
            theta (tensor_like or float): rotation angle

        Returns:
            tensor_like: eigenvalues

        **Example**

        >>> qml.RZ.compute_eigvals(torch.tensor(0.5))
        tensor([0.9689-0.2474j, 0.9689+0.2474j])
        """
        if qml.math.get_interface(theta) == "tensorflow":
            phase = qml.math.exp(-0.5j * qml.math.cast_like(theta, 1j))
            return qml.math.stack([phase, qml.math.conj(phase)], axis=-1)

        prefactors = qml.math.array([-0.5j, 0.5j], like=theta)
        if qml.math.ndim(theta) == 0:
            product = theta * prefactors
        else:
            product = qml.math.outer(theta, prefactors)
        return qml.math.exp(product)

    def adjoint(self):
        return RZ(-self.data[0], wires=self.wires)

    def pow(self, z):
        return [RZ(self.data[0] * z, wires=self.wires)]

    def _controlled(self, wire):
        return qml.CRZ(*self.parameters, wires=wire + self.wires)

    def simplify(self):
        theta = self.data[0] % (4 * np.pi)

        if _can_replace(theta, 0):
            return qml.Identity(wires=self.wires)

        return RZ(theta, wires=self.wires)

    def single_qubit_rot_angles(self):
        # RZ(\theta) = RZ(\theta) RY(0) RZ(0)
        return [self.data[0], 0.0, 0.0]


class PhaseShift(Operation):
    r"""
    Arbitrary single qubit local phase shift

    .. math:: R_\phi(\phi) = e^{i\phi/2}R_z(\phi) = \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\phi}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Number of dimensions per parameter: (0,)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R_\phi(\phi)) = \frac{1}{2}\left[f(R_\phi(\phi+\pi/2)) - f(R_\phi(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`R_{\phi}(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    basis = "Z"
    grad_method = "A"
    parameter_frequencies = [(1,)]

    def generator(self):
        return qml.Projector(np.array([1]), wires=self.wires)

    def __init__(self, phi, wires, do_queue=True, id=None):
        super().__init__(phi, wires=wires, do_queue=do_queue, id=id)

    def label(self, decimals=None, base_label=None, cache=None):
        return super().label(decimals=decimals, base_label=base_label or "Rϕ", cache=cache)

    @staticmethod
    def compute_matrix(phi):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.PhaseShift.matrix`


        Args:
            phi (tensor_like or float): phase shift

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.PhaseShift.compute_matrix(torch.tensor(0.5))
        tensor([[0.9689-0.2474j, 0.0000+0.0000j],
                [0.0000+0.0000j, 0.9689+0.2474j]])
        """
        if qml.math.get_interface(phi) == "tensorflow":
            p = qml.math.exp(1j * qml.math.cast_like(phi, 1j))
            ones = qml.math.ones_like(p)
            zeros = qml.math.zeros_like(p)

            return qml.math.stack([stack_last([ones, zeros]), stack_last([zeros, p])], axis=-2)

        signs = qml.math.array([0, 1], like=phi)
        arg = 1j * phi

        if qml.math.ndim(arg) == 0:
            return qml.math.diag(qml.math.exp(arg * signs))

        diags = qml.math.exp(qml.math.outer(arg, signs))
        return diags[:, :, np.newaxis] * qml.math.cast_like(qml.math.eye(2, like=diags), diags)

    @staticmethod
    def compute_eigvals(phi):  # pylint: disable=arguments-differ
        r"""Eigenvalues of the operator in the computational basis (static method).

        If :attr:`diagonalizing_gates` are specified and implement a unitary :math:`U^{\dagger}`,
        the operator can be reconstructed as

        .. math:: O = U \Sigma U^{\dagger},

        where :math:`\Sigma` is the diagonal matrix containing the eigenvalues.

        Otherwise, no particular order for the eigenvalues is guaranteed.

        .. seealso:: :meth:`~.PhaseShift.eigvals`


        Args:
            phi (tensor_like or float): phase shift

        Returns:
            tensor_like: eigenvalues

        **Example**

        >>> qml.PhaseShift.compute_eigvals(torch.tensor(0.5))
        tensor([1.0000+0.0000j, 0.8776+0.4794j])
        """
        if qml.math.get_interface(phi) == "tensorflow":
            phase = qml.math.exp(1j * qml.math.cast_like(phi, 1j))
            return stack_last([qml.math.ones_like(phase), phase])

        prefactors = qml.math.array([0, 1j], like=phi)
        if qml.math.ndim(phi) == 0:
            product = phi * prefactors
        else:
            product = qml.math.outer(phi, prefactors)
        return qml.math.exp(product)

    @staticmethod
    def compute_decomposition(phi, wires):
        r"""Representation of the operator as a product of other operators (static method). :

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.PhaseShift.decomposition`.

        Args:
            phi (float): rotation angle :math:`\phi`
            wires (Any, Wires): wires that the operator acts on

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.PhaseShift.compute_decomposition(1.234, wires=0)
        [RZ(1.234, wires=[0])]

        """
        return [RZ(phi, wires=wires)]

    def adjoint(self):
        return PhaseShift(-self.data[0], wires=self.wires)

    def pow(self, z):
        return [PhaseShift(self.data[0] * z, wires=self.wires)]

    def _controlled(self, wire):
        return qml.ControlledPhaseShift(*self.parameters, wires=wire + self.wires)

    def simplify(self):
        phi = self.data[0] % (2 * np.pi)

        if _can_replace(phi, 0):
            return qml.Identity(wires=self.wires)

        return PhaseShift(phi, wires=self.wires)

    def single_qubit_rot_angles(self):
        # PhaseShift(\theta) = RZ(\theta) RY(0) RZ(0)
        return [self.data[0], 0.0, 0.0]


class Rot(Operation):
    r"""
    Arbitrary single qubit rotation

    .. math::

        R(\phi,\theta,\omega) = RZ(\omega)RY(\theta)RZ(\phi)= \begin{bmatrix}
        e^{-i(\phi+\omega)/2}\cos(\theta/2) & -e^{i(\phi-\omega)/2}\sin(\theta/2) \\
        e^{-i(\phi-\omega)/2}\sin(\theta/2) & e^{i(\phi+\omega)/2}\cos(\theta/2)
        \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 3
    * Number of dimensions per parameter: (0, 0, 0)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(R(\phi, \theta, \omega)) = \frac{1}{2}\left[f(R(\phi+\pi/2, \theta, \omega)) - f(R(\phi-\pi/2, \theta, \omega))\right]`
      where :math:`f` is an expectation value depending on :math:`R(\phi, \theta, \omega)`.
      This gradient recipe applies for each angle argument :math:`\{\phi, \theta, \omega\}`.

    .. note::

        If the ``Rot`` gate is not supported on the targeted device, PennyLane
        will attempt to decompose the gate into :class:`~.RZ` and :class:`~.RY` gates.

    Args:
        phi (float): rotation angle :math:`\phi`
        theta (float): rotation angle :math:`\theta`
        omega (float): rotation angle :math:`\omega`
        wires (Any, Wires): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 3
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0, 0, 0)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = "A"
    parameter_frequencies = [(1,), (1,), (1,)]

    def __init__(self, phi, theta, omega, wires, do_queue=True, id=None):
        super().__init__(phi, theta, omega, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(phi, theta, omega):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.Rot.matrix`


        Args:
            phi (tensor_like or float): first rotation angle
            theta (tensor_like or float): second rotation angle
            omega (tensor_like or float): third rotation angle

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.Rot.compute_matrix(torch.tensor(0.1), torch.tensor(0.2), torch.tensor(0.3))
        tensor([[ 0.9752-0.1977j, -0.0993+0.0100j],
                [ 0.0993+0.0100j,  0.9752+0.1977j]])

        """
        # It might be that they are in different interfaces, e.g.,
        # Rot(0.2, 0.3, tf.Variable(0.5), wires=0)
        # So we need to make sure the matrix comes out having the right type
        interface = qml.math.get_interface(phi, theta, omega)

        c = qml.math.cos(theta / 2)
        s = qml.math.sin(theta / 2)

        # If anything is not tensorflow, it has to be casted and then
        if interface == "tensorflow":
            phi = qml.math.cast_like(qml.math.asarray(phi, like=interface), 1j)
            omega = qml.math.cast_like(qml.math.asarray(omega, like=interface), 1j)
            c = qml.math.cast_like(qml.math.asarray(c, like=interface), 1j)
            s = qml.math.cast_like(qml.math.asarray(s, like=interface), 1j)

        # The following variable is used to assert the all terms to be stacked have same shape
        one = qml.math.ones_like(phi) * qml.math.ones_like(omega)
        c = c * one
        s = s * one

        mat = [
            [
                qml.math.exp(-0.5j * (phi + omega)) * c,
                -qml.math.exp(0.5j * (phi - omega)) * s,
            ],
            [
                qml.math.exp(-0.5j * (phi - omega)) * s,
                qml.math.exp(0.5j * (phi + omega)) * c,
            ],
        ]

        return qml.math.stack([stack_last(row) for row in mat], axis=-2)

    @staticmethod
    def compute_decomposition(phi, theta, omega, wires):
        r"""Representation of the operator as a product of other operators (static method). :

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.Rot.decomposition`.

        Args:
            phi (float): rotation angle :math:`\phi`
            theta (float): rotation angle :math:`\theta`
            omega (float): rotation angle :math:`\omega`
            wires (Any, Wires): the wire the operation acts on

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.Rot.compute_decomposition(1.2, 2.3, 3.4, wires=0)
        [RZ(1.2, wires=[0]), RY(2.3, wires=[0]), RZ(3.4, wires=[0])]

        """
        decomp_ops = [
            RZ(phi, wires=wires),
            RY(theta, wires=wires),
            RZ(omega, wires=wires),
        ]
        return decomp_ops

    def adjoint(self):
        phi, theta, omega = self.parameters
        return Rot(-omega, -theta, -phi, wires=self.wires)

    def _controlled(self, wire):
        return qml.CRot(*self.parameters, wires=wire + self.wires)

    def single_qubit_rot_angles(self):
        return self.data

    def simplify(self):
        """Simplifies into single-rotation gates or a Hadamard if possible.

        >>> qml.Rot(np.pi / 2, 0.1, -np.pi / 2, wires=0).simplify()
        RX(0.1, wires=[0])
        >>> qml.Rot(np.pi, np.pi/2, 0, 0).simplify()
        Hadamard(wires=[0])

        """
        p0, p1, p2 = [p % (4 * np.pi) for p in self.data]

        if _can_replace(p0, 0) and _can_replace(p1, 0) and _can_replace(p2, 0):
            return qml.Identity(wires=self.wires)
        if _can_replace(p0, np.pi / 2) and _can_replace(p2, 7 * np.pi / 2):
            return RX(p1, wires=self.wires)
        if _can_replace(p0, 0) and _can_replace(p2, 0):
            return RY(p1, wires=self.wires)
        if _can_replace(p1, 0):
            return RZ((p0 + p2) % (4 * np.pi), wires=self.wires)
        if _can_replace(p0, np.pi) and _can_replace(p1, np.pi / 2) and _can_replace(p2, 0):
            return Hadamard(wires=self.wires)

        return Rot(p0, p1, p2, wires=self.wires)


class U1(Operation):
    r"""
    U1 gate.

    .. math:: U_1(\phi) = e^{i\phi/2}R_z(\phi) = \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\phi}
            \end{bmatrix}.

    .. note::

        The ``U1`` gate is an alias for the phase shift operation :class:`~.PhaseShift`.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Number of dimensions per parameter: (0,)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(U_1(\phi)) = \frac{1}{2}\left[f(U_1(\phi+\pi/2)) - f(U_1(\phi-\pi/2))\right]`
      where :math:`f` is an expectation value depending on :math:`U_1(\phi)`.

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = "A"
    parameter_frequencies = [(1,)]

    def generator(self):
        return qml.Projector(np.array([1]), wires=self.wires)

    def __init__(self, phi, wires, do_queue=True, id=None):
        super().__init__(phi, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(phi):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.U1.matrix`

        Args:
            phi (tensor_like or float): rotation angle

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.U1.compute_matrix(torch.tensor(0.5))
        tensor([[1.0000+0.0000j, 0.0000+0.0000j],
                [0.0000+0.0000j, 0.8776+0.4794j]])
        """
        if qml.math.get_interface(phi) == "tensorflow":
            phi = qml.math.cast_like(phi, 1j)
            fac = qml.math.cast_like([0, 1], 1j)
        else:
            fac = np.array([0, 1])

        fac = qml.math.convert_like(fac, phi)

        arg = 1j * phi
        if qml.math.ndim(arg) == 0:
            return qml.math.diag(qml.math.exp(arg * fac))

        diags = qml.math.exp(qml.math.outer(arg, fac))
        return diags[:, :, np.newaxis] * qml.math.cast_like(qml.math.eye(2, like=diags), diags)

    @staticmethod
    def compute_decomposition(phi, wires):
        r"""Representation of the operator as a product of other operators (static method). :

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.U1.decomposition`.

        Args:
            phi (float): rotation angle :math:`\phi`
            wires (Any, Wires): Wire that the operator acts on.

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.U1.compute_decomposition(1.234, wires=0)
        [PhaseShift(1.234, wires=[0])]

        """
        return [PhaseShift(phi, wires=wires)]

    def adjoint(self):
        return U1(-self.data[0], wires=self.wires)

    def pow(self, z):
        return [U1(self.data[0] * z, wires=self.wires)]

    def simplify(self):
        phi = self.data[0] % (2 * np.pi)

        if _can_replace(phi, 0):
            return qml.Identity(wires=self.wires)

        return U1(phi, wires=self.wires)


class U2(Operation):
    r"""
    U2 gate.

    .. math::

        U_2(\phi, \delta) = \frac{1}{\sqrt{2}}\begin{bmatrix} 1 & -\exp(i \delta)
        \\ \exp(i \phi) & \exp(i (\phi + \delta)) \end{bmatrix}

    The :math:`U_2` gate is related to the single-qubit rotation :math:`R` (:class:`Rot`) and the
    :math:`R_\phi` (:class:`PhaseShift`) gates via the following relation:

    .. math::

        U_2(\phi, \delta) = R_\phi(\phi+\delta) R(\delta,\pi/2,-\delta)

    .. note::

        If the ``U2`` gate is not supported on the targeted device, PennyLane
        will attempt to decompose the gate into :class:`~.Rot` and :class:`~.PhaseShift` gates.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 2
    * Number of dimensions per parameter: (0, 0)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(U_2(\phi, \delta)) = \frac{1}{2}\left[f(U_2(\phi+\pi/2, \delta)) - f(U_2(\phi-\pi/2, \delta))\right]`
      where :math:`f` is an expectation value depending on :math:`U_2(\phi, \delta)`.
      This gradient recipe applies for each angle argument :math:`\{\phi, \delta\}`.

    Args:
        phi (float): azimuthal angle :math:`\phi`
        delta (float): quantum phase :math:`\delta`
        wires (Sequence[int] or int): the subsystem the gate acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 2
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0, 0)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = "A"
    parameter_frequencies = [(1,), (1,)]

    def __init__(self, phi, delta, wires, do_queue=True, id=None):
        super().__init__(phi, delta, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(phi, delta):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.U2.matrix`

        Args:
            phi (tensor_like or float): azimuthal angle
            delta (tensor_like or float): quantum phase

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.U2.compute_matrix(torch.tensor(0.1), torch.tensor(0.2))
        tensor([[ 0.7071+0.0000j, -0.6930-0.1405j],
                [ 0.7036+0.0706j,  0.6755+0.2090j]])
        """
        interface = qml.math.get_interface(phi, delta)

        # If anything is not tensorflow, it has to be casted and then
        if interface == "tensorflow":
            phi = qml.math.cast_like(qml.math.asarray(phi, like=interface), 1j)
            delta = qml.math.cast_like(qml.math.asarray(delta, like=interface), 1j)

        one = qml.math.ones_like(phi) * qml.math.ones_like(delta)
        mat = [
            [one, -qml.math.exp(1j * delta) * one],
            [qml.math.exp(1j * phi) * one, qml.math.exp(1j * (phi + delta))],
        ]

        return qml.math.stack([stack_last(row) for row in mat], axis=-2) / np.sqrt(2)

    @staticmethod
    def compute_decomposition(phi, delta, wires):
        r"""Representation of the operator as a product of other operators (static method).

        .. math:: O = O_1 O_2 \dots O_n.

        .. seealso:: :meth:`~.U2.decomposition`.

        Args:
            phi (float): azimuthal angle :math:`\phi`
            delta (float): quantum phase :math:`\delta`
            wires (Iterable, Wires): the subsystem the gate acts on

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.U2.compute_decomposition(1.23, 2.34, wires=0)
        [Rot(2.34, 1.5707963267948966, -2.34, wires=[0]),
        PhaseShift(2.34, wires=[0]),
        PhaseShift(1.23, wires=[0])]

        """
        pi_half = qml.math.ones_like(delta) * (np.pi / 2)
        decomp_ops = [
            Rot(delta, pi_half, -delta, wires=wires),
            PhaseShift(delta, wires=wires),
            PhaseShift(phi, wires=wires),
        ]
        return decomp_ops

    def adjoint(self):
        phi, delta = self.parameters
        new_delta = qml.math.mod((np.pi - phi), (2 * np.pi))
        new_phi = qml.math.mod((np.pi - delta), (2 * np.pi))
        return U2(new_phi, new_delta, wires=self.wires)

    def simplify(self):
        """Simplifies the gate into RX or RY gates if possible."""
        wires = self.wires

        phi, delta = [p % (2 * np.pi) for p in self.data]

        if _can_replace(delta, 0) and _can_replace(phi, 0):
            return RY(np.pi / 2, wires=wires)
        if _can_replace(delta, np.pi / 2) and _can_replace(phi, 3 * np.pi / 2):
            return RX(np.pi / 2, wires=wires)
        if _can_replace(delta, 3 * np.pi / 2) and _can_replace(phi, np.pi / 2):
            return RX(3 * np.pi / 2, wires=wires)

        return U2(phi, delta, wires=wires)


class U3(Operation):
    r"""
    Arbitrary single qubit unitary.

    .. math::

        U_3(\theta, \phi, \delta) = \begin{bmatrix} \cos(\theta/2) & -\exp(i \delta)\sin(\theta/2) \\
        \exp(i \phi)\sin(\theta/2) & \exp(i (\phi + \delta))\cos(\theta/2) \end{bmatrix}

    The :math:`U_3` gate is related to the single-qubit rotation :math:`R` (:class:`Rot`) and the
    :math:`R_\phi` (:class:`PhaseShift`) gates via the following relation:

    .. math::

        U_3(\theta, \phi, \delta) = R_\phi(\phi+\delta) R(\delta,\theta,-\delta)

    .. note::

        If the ``U3`` gate is not supported on the targeted device, PennyLane
        will attempt to decompose the gate into :class:`~.PhaseShift` and :class:`~.Rot` gates.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 3
    * Number of dimensions per parameter: (0, 0, 0)
    * Gradient recipe: :math:`\frac{d}{d\phi}f(U_3(\theta, \phi, \delta)) = \frac{1}{2}\left[f(U_3(\theta+\pi/2, \phi, \delta)) - f(U_3(\theta-\pi/2, \phi, \delta))\right]`
      where :math:`f` is an expectation value depending on :math:`U_3(\theta, \phi, \delta)`.
      This gradient recipe applies for each angle argument :math:`\{\theta, \phi, \delta\}`.

    Args:
        theta (float): polar angle :math:`\theta`
        phi (float): azimuthal angle :math:`\phi`
        delta (float): quantum phase :math:`\delta`
        wires (Sequence[int] or int): the subsystem the gate acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)
    """
    num_wires = 1
    num_params = 3
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (0, 0, 0)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = "A"
    parameter_frequencies = [(1,), (1,), (1,)]

    def __init__(self, theta, phi, delta, wires, do_queue=True, id=None):
        super().__init__(theta, phi, delta, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(theta, phi, delta):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.U3.matrix`

        Args:
            theta (tensor_like or float): polar angle
            phi (tensor_like or float): azimuthal angle
            delta (tensor_like or float): quantum phase

        Returns:
            tensor_like: canonical matrix

        **Example**

        >>> qml.U3.compute_matrix(torch.tensor(0.1), torch.tensor(0.2), torch.tensor(0.3))
        tensor([[ 0.9988+0.0000j, -0.0477-0.0148j],
                [ 0.0490+0.0099j,  0.8765+0.4788j]])

        """
        # It might be that they are in different interfaces, e.g.,
        # U3(0.2, 0.3, tf.Variable(0.5), wires=0)
        # So we need to make sure the matrix comes out having the right type
        interface = qml.math.get_interface(theta, phi, delta)

        c = qml.math.cos(theta / 2)
        s = qml.math.sin(theta / 2)

        # If anything is not tensorflow, it has to be casted and then
        if interface == "tensorflow":
            phi = qml.math.cast_like(qml.math.asarray(phi, like=interface), 1j)
            delta = qml.math.cast_like(qml.math.asarray(delta, like=interface), 1j)
            c = qml.math.cast_like(qml.math.asarray(c, like=interface), 1j)
            s = qml.math.cast_like(qml.math.asarray(s, like=interface), 1j)

        # The following variable is used to assert the all terms to be stacked have same shape
        one = qml.math.ones_like(phi) * qml.math.ones_like(delta)
        c = c * one
        s = s * one

        mat = [
            [c, -s * qml.math.exp(1j * delta)],
            [s * qml.math.exp(1j * phi), c * qml.math.exp(1j * (phi + delta))],
        ]

        return qml.math.stack([stack_last(row) for row in mat], axis=-2)

    @staticmethod
    def compute_decomposition(theta, phi, delta, wires):
        r"""Representation of the operator as a product of other operators (static method).

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.U3.decomposition`.

        Args:
            theta (float): polar angle :math:`\theta`
            phi (float): azimuthal angle :math:`\phi`
            delta (float): quantum phase :math:`\delta`
            wires (Iterable, Wires): the subsystem the gate acts on

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.U3.compute_decomposition(1.23, 2.34, 3.45, wires=0)
        [Rot(3.45, 1.23, -3.45, wires=[0]),
        PhaseShift(3.45, wires=[0]),
        PhaseShift(2.34, wires=[0])]

        """
        decomp_ops = [
            Rot(delta, theta, -delta, wires=wires),
            PhaseShift(delta, wires=wires),
            PhaseShift(phi, wires=wires),
        ]
        return decomp_ops

    def adjoint(self):
        theta, phi, delta = self.parameters
        new_delta = qml.math.mod((np.pi - phi), (2 * np.pi))
        new_phi = qml.math.mod((np.pi - delta), (2 * np.pi))
        return U3(theta, new_phi, new_delta, wires=self.wires)

    def simplify(self):
        """Simplifies into :class:`~.RX`, :class:`~.RY`, or :class:`~.PhaseShift` gates
        if possible.

        >>> qml.U3(0.1, 0, 0, wires=0).simplify()
        RY(0.1, wires=[0])

        """
        wires = self.wires
        params = self.parameters

        p0 = params[0] % (4 * np.pi)
        p1, p2 = [p % (2 * np.pi) for p in params[1:]]

        if _can_replace(p0, 0) and _can_replace(p1, 0) and _can_replace(p2, 0):
            return qml.Identity(wires=wires)
        if _can_replace(p0, 0) and not _can_replace(p1, 0) and _can_replace(p2, 0):
            return PhaseShift(p1, wires=wires)
        if (
            _can_replace(p2, np.pi / 2)
            and _can_replace(p1, 3 * np.pi / 2)
            and not _can_replace(p0, 0)
        ):
            return RX(p0, wires=wires)
        if not _can_replace(p0, 0) and _can_replace(p1, 0) and _can_replace(p2, 0):
            return RY(p0, wires=wires)

        return U3(p0, p1, p2, wires=wires)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This submodule contains the operation SpecialUnitary and
its utility functions.
"""
# pylint: disable=arguments-differ, import-outside-toplevel
from functools import lru_cache, reduce
from itertools import product

import numpy as np
import pennylane as qml
from pennylane.operation import AnyWires, Operation
from pennylane.ops.qubit.parametric_ops_multi_qubit import PauliRot

_pauli_matrices = np.array(
    [[[1, 0], [0, 1]], [[0, 1], [1, 0]], [[0, -1j], [1j, 0]], [[1, 0], [0, -1]]]
)
"""Single-qubit Paulis:    I                 X                   Y                  Z"""

_pauli_letters = "IXYZ"
"""Single-qubit Pauli letters that make up Pauli words."""


@lru_cache
def pauli_basis_matrices(num_wires):
    r"""Compute all elements of the Pauli basis of the Lie algebra :math:`\mathfrak{su}(N)`
    as a single, dense tensor.

    Args:
        num_wires (int): The number of wires on which the associated Pauli group acts.

    Returns:
        ndarray: All Pauli basis elements of :math:`\mathfrak{su}(N)`.

    The basis has :math:`d=4^n-1` elements for :math:`n` qubits, yielding an output tensor
    with shape ``(4**num_wires-1, 2**num_wires, 2**num_wires)`` and :math:`16^n-4^n` entries.
    The identity Pauli word :math:`I^{\otimes n}` does not belong to :math:`\mathfrak{su}(N)`
    and therefore is not included.

    The basis elements are ordered (choose the description that suits you most)

      - lexicographically.

      - such that the term acting on the last qubit changes fastest, the one acting on the first
        qubit changes slowest when iterating through the output.

      - such that the basis index, written in base :math:`4`, contains the indices for the list
        ``["I", "X", "Y", "Z"]``, in the order of the qubits

      - such that for three qubits, the first basis elements are the Pauli words
        ``"IIX", ""IIY", "IIZ", "IXI", "IXX", "IXY", "IXZ", "IYI"...``

    .. note::

        Note that this method internally handles a complex-valued tensor of size
        ``(4**num_wires, 2**num_wires, 2**num_wires)``, which requires at least
        ``4 ** (2 * num_wires - 13)`` GB of memory (at default precision).

    **Example**

    >>> pauli_basis_matrices(1)
    array([[[ 0.+0.j,  1.+0.j],
            [ 1.+0.j,  0.+0.j]],
           [[ 0.+0.j, -0.-1.j],
            [ 0.+1.j,  0.+0.j]],
           [[ 1.+0.j,  0.+0.j],
            [ 0.+0.j, -1.+0.j]]])

    >>> pauli_basis_matrices(3).shape
    (63, 8, 8)
    """
    if num_wires < 1:
        raise ValueError(f"Require at least one wire, got {num_wires}.")
    if num_wires > 7:
        raise ValueError(
            f"Creating the Pauli basis tensor for more than 7 wires (got {num_wires}) is deactivated."
        )
    return reduce(np.kron, (_pauli_matrices for _ in range(num_wires)))[1:]


@lru_cache
def pauli_basis_strings(num_wires):
    r"""Compute all :math:`n`-qubit Pauli words except ``"I"*num_wires``,
    corresponding to the Pauli basis of the Lie algebra :math:`\mathfrak{su}(N)`.

    Args:
        num_wires (int): The number of wires, or number of letters per word.

    Returns:
        list[str]: All Pauli words on ``num_wires`` qubits, except from the identity.

    There are :math:`d=4^n-1` Pauli words that are not the identity. They are ordered
    (choose the description that suits you most)

      - lexicographically.

      - such that the term acting on the last qubit changes fastest, the one acting on the first
        qubit changes slowest when iterating through the output.

      - such that the basis index, written in base :math:`4`, contains the indices for the list
        ``["I", "X", "Y", "Z"]``, in the order of the qubits

      - such that for three qubits, the first Pauli words are
        ``"IIX", ""IIY", "IIZ", "IXI", "IXX", "IXY", "IXZ", "IYI"...``

    **Example**

    >>> pauli_basis_strings(1)
    ['X', 'Y', 'Z']
    >>> len(pauli_basis_strings(3))
    63
    """
    return ["".join(letters) for letters in product(_pauli_letters, repeat=num_wires)][1:]


class SpecialUnitary(Operation):
    r"""Gate from the group :math:`SU(N)` with :math:`N=2^n` for :math:`n` qubits.

    We use the following parametrization of a special unitary operator:

    .. math::

        U(\bm{\theta}) &= \exp\{A(\bm{\theta})\}\\
        A(\bm{\theta}) &= \sum_{m=1}^d i \theta_m P_m\\
        P_m &\in \{I, X, Y, Z\}^{\otimes n} \setminus \{I^{\otimes n}\}

    This means, :math:`U(\bm{\theta})` is the exponential of the operator :math:`A(\bm{\theta})`,
    which in turn is a linear combination of Pauli words with coefficients :math:`i\bm{\theta}`
    and satisfies :math:`A(\bm{\theta})^\dagger=-A(\bm{\theta})` (it is *skew-Hermitian*).
    Note that this gate takes an exponential number :math:`d=4^n-1` of parameters.
    See below for more theoretical background and details regarding differentiability.

    **Details:**

    * Number of wires: Any
    * Number of parameters: 1
    * Number of dimensions per parameter: (1,)
    * Gradient recipe:

    .. math::

        \frac{\partial}{\partial\theta_\ell} f(U(\bm{\theta})) &= \sum_{m=1}^d 2i\omega_{\ell m}
        \frac{\mathrm{d}}{\mathrm{d} t}
        f\left(\exp\left\{-i\frac{t}{2}G_m\right\} U(\bm{\theta})\right)\\
        &= \sum_{m=1}^d i\omega_{\ell m}
        \left[
        f\left(\exp\left\{-i\frac{\pi}{4}G_m\right\} U(\bm{\theta})\right)
        -f\left(\exp\left\{i\frac{\pi}{4}G_m\right\} U(\bm{\theta})\right)
        \right]

    where :math:`f` is an expectation value depending on :math:`U(\bm{\theta})` and the derivative
    of the Pauli rotation gates :math:`\exp\left\{-i\frac{t}{2}G_m\right\}` follows from
    the two-term parameter-shift rule (also see: :class:`~.PauliRot`). For details on
    the gradient recipe, also consider the theoretical background section below.

    Args:
        theta (tensor_like): Pauli coordinates of the exponent :math:`A(\bm{\theta})`.
            See details below for the order of the Pauli words.
        wires (Sequence[int] or int): The wire(s) the operation acts on
        do_queue (bool): Indicates whether the operator should be
            immediately pushed into the Operator queue (optional)
        id (str or None): String representing the operation (optional)

    Raises:
        ValueError: If the shape of the input does not match the Lie algebra
            dimension :math:`d=4^n-1` for :math:`n` wires.

    .. note::

        This operation should only be used to parametrize special unitaries that
        can not easily be represented by other operations, as it incurs
        computational cost that scale exponentially with the wires it acts on.

    The parameter ``theta`` refers to all Pauli words (except for the identity) in
    lexicographical order, which looks like the following for one and two qubits:

    >>> qml.ops.qubit.special_unitary.pauli_basis_strings(1) # 4**1-1 = 3 Pauli words
    ['X', 'Y', 'Z']
    >>> qml.ops.qubit.special_unitary.pauli_basis_strings(2) # 4**2-1 = 15 Pauli words
    ['IX', 'IY', 'IZ', 'XI', 'XX', 'XY', 'XZ', 'YI', 'YX', 'YY', 'YZ', 'ZI', 'ZX', 'ZY', 'ZZ']

    .. warning::

        This operation only is differentiable when using the JAX, Torch or Tensorflow
        interfaces, even when using hardware-compatible differentiation techniques like
        the parameter-shift rule.


    .. warning::

        This operation supports broadcasting and hardware-compatible differentiation
        techniques like the parameter-shift rule. However, the two features can not be
        used simultaneously.

    **Examples**

    Simple examples of this operation are single-qubit Pauli rotation gates, which can be
    created by setting all but one parameter :math:`\theta_m` to zero:

    >>> x = 0.412
    >>> theta = x * np.array([1, 0, 0]) # The first entry belongs to the Pauli word "X"
    >>> su = qml.SpecialUnitary(theta, wires=0)
    >>> prot = qml.PauliRot(-2 * x, "X", wires=0) # PauliRot introduces a prefactor -0.5
    >>> rx = qml.RX(-2 * x, 0) # RX introduces a prefactor -0.5
    >>> qml.math.allclose(su.matrix(), prot.matrix())
    True
    >>> qml.math.allclose(su.matrix(), rx.matrix())
    True

    Note that for specific operations like the ``RX`` rotation gate above, it is
    strongly recommended to use the specialized implementation ``qml.RX`` rather
    than ``PauliRot`` or ``SpecialUnitary``.
    However, ``SpecialUnitary`` gates go beyond such rotations: Multiple Pauli words
    can be activated simultaneously, giving access to more complex operations.
    For two qubits, this could look like this:

    >>> wires = [0, 1]
    # Activating the Pauli words ["IY", "IZ", "XX", "XY", "YY", "YZ", "ZY", "ZZ"]
    >>> theta = 0.3 * np.array([0, 1, 2, 0, -1, 1, 0, 0, 0, 1, 1, 1, 0, 0, -1])
    >>> len(theta) == 4 ** len(wires) - 1 # theta contains one parameter per Pauli word
    True
    >>> su = qml.SpecialUnitary(theta, wires=wires)
    >>> su.matrix()
    array([[ 0.56397118+0.52139241j,  0.30652227+0.02438052j,
             0.13555302+0.22630716j,  0.0689876 -0.49110826j],
           [-0.15454843+0.00998377j,  0.88294943+0.01496327j,
            -0.25396275-0.10785888j, -0.26041566+0.22857073j],
           [-0.2876174 -0.2443733j ,  0.25423439+0.05896445j,
             0.71621665+0.50686226j,  0.1380692 +0.02252197j],
           [-0.34495668-0.35307844j,  0.10817019-0.21404059j,
            -0.29040522+0.00830631j,  0.15015337-0.76933485j]])

    The ``SpecialUnitary`` operation also can be differentiated with hardware-compatible
    differentiation techniques if the JAX, Torch or Tensorflow interface is used.
    See the theoretical background section below for details.

    .. details::
        :title: Theoretical background
        :href: theory

        We consider special unitaries parametrized in the following way:

        .. math::

            U(\bm{\theta}) &= \exp\{A(\bm{\theta})\}\\
            A(\bm{\theta}) &= \sum_{m=1}^d i \theta_m G_m\\
            G_m &\in \mathcal{P^{(n)}}=\{I, X, Y, Z\}^{\otimes n} \setminus \{I^{\otimes n}\}

        where :math:`n` is the number of qubits and :math:`\theta_m` are :math:`d=4^n-1`
        real-valued parameters. This parametrization allows us to express any special unitary
        for the given number of qubits.

        Note that this differs from a sequence of all possible Pauli rotation gates because
        for non-commuting Pauli words :math:`G_1, G_2` we have
        :math:`\exp\{i\theta_1G_1\}\exp\{i\theta_2G_2\}\neq \exp\{i(\theta_1G_1+\theta_2G_2)\}`.

        **Differentiation**

        The partial derivatives of :math:`U(\bm{\theta})` above can be expressed as

        .. math::

            \frac{\partial}{\partial \theta_\ell} U(\bm{\theta}) &= U(\bm{\theta})
            \frac{\mathrm{d}}{\mathrm{d}t}\exp\left(t\Omega_\ell(\bm{\theta})\right)\large|_{t=0}\\
            &=U(\bm{\theta})\Omega_\ell(\bm{\theta})

        where :math:`\Omega_\ell(\bm{\theta})` is the *effective generator* belonging to the partial
        derivative :math:`\partial_\ell U(\bm{\theta})` at the parameters :math:`\bm{\theta}`.
        It can be computed via

        .. math::

            \Omega_\ell(\bm{\theta}) = U(\bm{\theta})^\dagger
            \left(\frac{\partial}{\partial \theta_\ell}\mathfrak{Re}[U(\bm{\theta})]
            +i\frac{\partial}{\partial \theta_\ell}\mathfrak{Im}[U(\bm{\theta})]\right)

        where we may compute the derivatives of the real and imaginary parts of :math:`U(\bm{\theta})`
        using auto-differentiation.

        Each effective generator :math:`\Omega_\ell(\bm{\theta})` that reproduces a partial derivative
        can be decomposed in the Pauli basis of :math:`\mathfrak{su}(N)` via

        .. math::

            \Omega_\ell(\bm{\theta}) = \sum_{m=1}^d \omega_{\ell m}(\bm{\theta}) G_m.

        As the Pauli words are orthonormal with respect to the
        `trace, or Frobenius, inner product <https://en.wikipedia.org/wiki/Frobenius_inner_product>`__
        (rescaled by :math:`2^{-n}`), we can compute the coefficients using this
        inner product (:math:`G_m` is Hermitian, so we skip the adjoint :math:`{}^\dagger`):

        .. math::

            \omega_{\ell m}(\bm{\theta}) = \frac{1}{2^n}\operatorname{tr}
            \left\{G_m \Omega_\ell(\bm{\theta}) \right\}

        The coefficients satisfy :math:`\omega_{\ell m}^\ast=-\omega_{\ell m}` because
        :math:`\Omega_\ell(\bm{\theta})` is skew-Hermitian. Therefore they are purely imaginary.

        Now we turn to the derivative of an expectation value-based function which uses a circuit
        with a ``SpecialUnitary`` operation. Absorbing the remaining circuit into the
        quantum state :math:`\rho` and the observable :math:`H`, this can be written as

        .. math::

            f(U(\bm{\theta})) &= \operatorname{Tr}\left\{H U(\bm{\theta})\rho
            U^\dagger(\bm{\theta})\right\}\\
            \partial_\ell f(U(\bm{\theta})) &= \operatorname{Tr}\left\{H U(\bm{\theta})
            [\Omega_\ell(\bm{\theta}),\rho] U^\dagger(\bm{\theta})\right\}

        Inserting the decomposition for the effective generator from above, we may rewrite this as
        a combination of derivatives of Pauli rotation gates:

        .. math::

            \partial_\ell f(U(\bm{\theta}))
            &= \operatorname{Tr}\left\{H U(\bm{\theta})
            \left[\sum_{m=1}^d \omega_{\ell m}(\bm{\theta}) G_m,\rho\right]
            U^\dagger(\bm{\theta})\right\}\\
            &= \sum_{m=1}^d 2i\omega_{\ell m}(\bm{\theta})
            \frac{\mathrm{d}}{\mathrm{d}t}f\left(R_{G_m}(t)U(\bm{\theta})\right)\bigg|_{t=0}.

        Here we abbreviated a Pauli rotation gate as :math:`R_{G_m}(t) = \exp\{-i\frac{t}{2} G_m\}`.
        As all partial derivatives are combinations of these Pauli rotation derivatives, we may
        write the gradient of :math:`f` as

        .. math::

            \nabla f(U(\bm{\theta})) = \overline{\omega}(\bm{\theta}) \cdot \bm{\mathrm{d}f}

        where we defined the matrix :math:`\overline{\omega_{\ell m}}=2i\omega_{\ell m}` and
        the vector of Pauli rotation derivatives
        :math:`\mathrm{d}f_m=\frac{\mathrm{d}}{\mathrm{d}t}
        f\left(R_{G_m}(t)U(\bm{\theta})\right)\bigg|_{t=0}`.
        These derivatives can be computed with the
        standard parameter-shift rule because Pauli words satisfy the condition :math:`G_m^2=1`
        (see e.g. `Mitarai et al. (2018) <https://arxiv.org/abs/1803.00745>`_).
        Provided that we can implement the ``SpecialUnitary`` gate
        itself, this allows us to compute :math:`\bm{\mathrm{d}f}` in a hardware-compatible
        manner using :math:`2d` quantum circuits.

        Note that for ``SpecialUnitary`` we frequently handle objects that have one or
        multiple dimensions of exponentially large size :math:`d=4^n-1`, and that the number
        of quantum circuits for the differentiation scales accordingly. This strongly affects
        the number of qubits to which we can apply a ``SpecialUnitary`` gate in practice.

    """
    num_wires = AnyWires
    """int: Number of wires that the operator acts on."""

    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (1,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = None
    """Gradient computation method."""

    def __init__(self, theta, wires, do_queue=True, id=None):
        num_wires = 1 if isinstance(wires, int) else len(wires)
        self.hyperparameters["num_wires"] = num_wires
        theta_shape = qml.math.shape(theta)
        expected_dim = 4**num_wires - 1

        if len(theta_shape) not in {1, 2}:
            raise ValueError(
                "Expected the parameters to have one or two dimensions without or with "
                f"broadcasting, respectively. The parameters have shape {theta_shape}"
            )

        if theta_shape[-1] != expected_dim:
            raise ValueError(
                f"Expected the parameters to have shape ({expected_dim},) or (batch_size, "
                f"{expected_dim}). The parameters have shape {theta_shape}"
            )

        super().__init__(theta, wires=wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_matrix(theta, num_wires):
        r"""Representation of the operator as a canonical matrix in the computational basis
        (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.SpecialUnitary.matrix`

        Args:
            theta (tensor_like): Pauli coordinates of the exponent :math:`A(\theta)`.
            num_wires (int): The number of wires the matrix acts on.

        Returns:
            tensor_like: canonical matrix of the special unitary corresponding to ``theta``. It
                has the shape ``(2**num_wires, 2**num_wires)``.

        Compute the matrix of an element in SU(N), given by the Pauli basis coordinated
        of the associated Lie algebra element.
        The :math:`4^n-1` Pauli basis elements of the Lie algebra :math:`\mathfrak{su}(N)`
        for :math:`n` qubits are
        :math:`P_m\in\{I, X, Y, Z\}^{\otimes n}\setminus\{I^{\otimes n}\}`, and the special
        unitary matrix is computed as

        .. math::

            U(\theta) = \exp(i\sum_{m=1}^d \theta_m P_m)

        See the main class documentation above for the ordering of Pauli words.

        .. note::

            Note that this method internally handles a complex-valued tensor of size
            ``(4**num_wires, 2**num_wires, 2**num_wires)``, which requires at least
            ``4 ** (2 * num_wires - 13)`` GB of memory (at default precision).

        **Example**

        >>> theta = np.array([0.5, 0.1, -0.3])
        >>> qml.SpecialUnitary.compute_matrix(theta, num_wires=1)
        array([[ 0.83004499-0.28280371j,  0.0942679 +0.47133952j],
               [-0.0942679 +0.47133952j,  0.83004499+0.28280371j]])
        """
        interface = qml.math.get_interface(theta)
        if interface == "tensorflow":
            theta = qml.math.cast_like(theta, 1j)
        if num_wires > 5:
            matrices = product(_pauli_matrices, repeat=num_wires)
            # Drop the identity from the generator of matrices
            _ = next(matrices)
            A = sum(t * reduce(np.kron, letters) for t, letters in zip(theta, matrices))
        else:
            A = qml.math.tensordot(theta, pauli_basis_matrices(num_wires), axes=[[-1], [0]])
        if interface == "jax" and qml.math.ndim(theta) > 1:
            # jax.numpy.expm does not support broadcasting
            return qml.math.stack([qml.math.expm(1j * _A) for _A in A])
        return qml.math.expm(1j * A)

    def get_one_parameter_generators(self, interface=None):
        r"""Compute the generators of one-parameter groups that reproduce
        the partial derivatives of a special unitary gate.

        Args:
            interface (str): The auto-differentiation framework to be used for the
                computation. Has to be one of ``["jax", "tensorflow", "tf", "torch"]``.

        Raises:
            NotImplementedError: If the chosen interface is ``"autograd"``. Autograd
                does not support differentiation of ``linalg.expm``.
            ValueError: If the chosen interface is not supported.

        Returns:
            tensor_like: The generators for one-parameter groups that reproduce the
            partial derivatives of the special unitary gate.
            There are :math:`d=4^n-1` generators for :math:`n` qubits, so that the
            output shape is ``(4**num_wires-1, 2**num_wires, 2**num_wires)``.

        Consider a special unitary gate parametrized in the following way:

        .. math::

            U(\theta) &= \exp\{A(\theta)\}\\
            A(\theta) &= \sum_{m=1}^d i \theta_m P_m\\
            P_m &\in \{I, X, Y, Z\}^{\otimes n} \setminus \{I^{\otimes n}\}

        Then the partial derivatives of the gate can be shown to be given by

        .. math::

            \frac{\partial}{\partial \theta_\ell} U(\theta) = U(\theta)
            \frac{\mathrm{d}}{\mathrm{d}t}\exp\left(t\Omega_\ell(\theta)\right)\large|_{t=0}

        where :math:`\Omega_\ell(\theta)` is the one-parameter generator belonging to the partial
        derivative :math:`\partial_\ell U(\theta)` at the parameters :math:`\theta`.
        It can be computed via

        .. math::

            \Omega_\ell(\theta) = U(\theta)^\dagger
            \left(\frac{\partial}{\partial \theta_\ell}\mathfrak{Re}[U(\theta)]
            +i\frac{\partial}{\partial \theta_\ell}\mathfrak{Im}[U(\theta)]\right)

        where we may compute the derivatives :math:`\frac{\partial}{\partial \theta_\ell} U(\theta)` using auto-differentiation.

        .. warning::

            An auto-differentiation framework is required for this function.
            The matrix exponential is not differentiable in Autograd. Therefore this function
            only supports JAX, Torch and Tensorflow.

        """
        theta = self.data[0]
        if len(qml.math.shape(theta)) > 1:
            raise ValueError("Broadcasting is not supported.")

        num_wires = self.hyperparameters["num_wires"]

        def split_matrix(theta):
            """Compute the real and imaginary parts of the special unitary matrix."""
            mat = self.compute_matrix(theta, num_wires)
            return qml.math.real(mat), qml.math.imag(mat)

        if interface == "jax":
            import jax

            theta = qml.math.cast_like(theta, 1j)
            # These lines compute the Jacobian of compute_matrix every time -> to be optimized
            jac = jax.jacobian(self.compute_matrix, argnums=0, holomorphic=True)(theta, num_wires)

        elif interface == "torch":
            import torch

            rjac, ijac = torch.autograd.functional.jacobian(split_matrix, theta)
            jac = rjac + 1j * ijac

        elif interface in ("tensorflow", "tf"):
            import tensorflow as tf

            with tf.GradientTape(persistent=True) as tape:
                mats = qml.math.stack(split_matrix(theta))

            rjac, ijac = tape.jacobian(mats, theta)
            jac = qml.math.cast_like(rjac, 1j) + 1j * qml.math.cast_like(ijac, 1j)

        elif interface == "autograd":
            # TODO check whether we can add support for Autograd using eigenvalue decomposition
            raise NotImplementedError(
                "The matrix exponential expm is not differentiable in Autograd."
            )

        else:
            raise ValueError(f"The interface {interface} is not supported.")

        # Compute the Omegas from the Jacobian. The adjoint of U(theta) is realized via -theta
        U_dagger = self.compute_matrix(-qml.math.detach(theta), num_wires)
        # After contracting, move the parameter derivative axis to the first position
        return qml.math.transpose(qml.math.tensordot(U_dagger, jac, axes=[[1], [0]]), [2, 0, 1])

    def get_one_parameter_coeffs(self, interface):
        r"""Compute the Pauli basis coefficients of the generators of one-parameter groups
        that reproduce the partial derivatives of a special unitary gate.

        Args:
            interface (str): The auto-differentiation framework to be used for the
                computation.

        Returns:
            tensor_like: The Pauli basis coefficients of the effective generators
            that reproduce the partial derivatives of the special unitary gate
            defined by ``theta``. There are :math:`d=4^n-1` generators for
            :math:`n` qubits and :math:`d` Pauli coefficients per generator, so that the
            output shape is ``(4**num_wires-1, 4**num_wires-1)``.

        Given a generator :math:`\Omega` of a one-parameter group that
        reproduces a partial derivative of a special unitary gate, it can be decomposed in
        the Pauli basis of :math:`\mathfrak{su}(N)` via

        .. math::

            \Omega = \sum_{m=1}^d \omega_m P_m

        where :math:`d=4^n-1` is the size of the basis for :math:`n` qubits and :math:`P_m` are the
        Pauli words making up the basis. As the Pauli words are orthonormal with respect to the
        `trace or Frobenius inner product <https://en.wikipedia.org/wiki/Frobenius_inner_product>`__
        (rescaled by :math:`2^n`), we can compute the coefficients using this
        inner product (:math:`P_m` is Hermitian, so we skip the adjoint :math:`{}^\dagger`):

        .. math::

            \omega_m = \frac{1}{2^n}\operatorname{tr}\left[P_m \Omega \right]

        The coefficients satisfy :math:`\omega_m^\ast=-\omega_m` because :math:`\Omega` is
        skew-Hermitian. Therefore they are purely imaginary.

        .. warning::

            An auto-differentiation framework is required by this function.
            The matrix exponential is not differentiable in Autograd. Therefore this function
            only supports JAX, Torch and Tensorflow.

        .. seealso:: :meth:`~.SpecialUnitary.get_one_parameter_generators`

        """
        num_wires = self.hyperparameters["num_wires"]
        basis = pauli_basis_matrices(num_wires)
        generators = self.get_one_parameter_generators(interface)
        return qml.math.tensordot(basis, generators, axes=[[1, 2], [2, 1]]) / 2**num_wires

    def decomposition(self):
        r"""Representation of the operator as a product of other operators.

        .. math:: O = O_1 O_2 \dots O_n

        This ``Operation`` is decomposed into the corresponding ``QubitUnitary``.

        Returns:
            list[Operator]: decomposition of the operator

        **Example:**

        >>> theta = np.array([0.5, 0.1, -0.3])
        >>> qml.SpecialUnitary(theta, wires=[0]).decomposition()
        [QubitUnitary(array([[ 0.83004499-0.28280371j,  0.0942679 +0.47133952j],
            [-0.0942679 +0.47133952j,  0.83004499+0.28280371j]]), wires=[0])]
        """
        theta = self.data[0]
        if qml.math.requires_grad(theta):
            interface = qml.math.get_interface(theta)
            # Get all Pauli words for the basis of the Lie algebra for this gate
            words = pauli_basis_strings(self.hyperparameters["num_wires"])

            # Compute the linear map that transforms between the Pauli basis and effective generators
            # Consider the mathematical derivation for the prefactor 2j
            omega = qml.math.real(2j * self.get_one_parameter_coeffs(interface))

            # Create zero parameters for each Pauli rotation gate that take over the trace of theta
            detached_theta = qml.math.detach(theta)
            zeros = theta - detached_theta
            # Apply the linear map omega to the zeros to create the correct preprocessing Jacobian
            zeros = qml.math.tensordot(omega, zeros, axes=[[1], [0]])

            # Apply Pauli rotations that yield the Pauli basis derivatives
            paulirots = [
                TmpPauliRot(zero, word, wires=self.wires, id="SU(N) byproduct")
                for zero, word in zip(zeros, words)
            ]
            return paulirots + [SpecialUnitary(detached_theta, wires=self.wires)]

        return [qml.QubitUnitary(self.matrix(), wires=self.wires)]

    def adjoint(self):
        return SpecialUnitary(-self.data[0], wires=self.wires)


class TmpPauliRot(PauliRot):
    r"""A custom version of ``PauliRot`` that is inserted with rotation angle zero when
    decomposing ``SpecialUnitary``. The differentiation logic makes use of the gradient
    recipe of ``PauliRot``, but deactivates the matrix property so that a decomposition
    of differentiated tapes is forced. During this decomposition, this private operation
    removes itself if its angle remained at zero.

    For details see :class:`~.PauliRot`.

    .. warning::

        Do not add this operation to the supported operations of any device.
        Wrong results and/or severe performance degradations may result.
    """

    # Deactivate the matrix property of qml.PauliRot in order to force decomposition
    has_matrix = False

    @staticmethod
    def compute_decomposition(theta, wires, pauli_word):
        r"""Representation of the operator as a product of other operators (static method). :

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.TmpPauliRot.decomposition`.

        Args:
            theta (float): rotation angle :math:`\theta`
            wires (Iterable, Wires): the wires the operation acts on
            pauli_word (string): the Pauli word defining the rotation

        Returns:
            list[Operator]: decomposition into an empty list of operations for
            vanishing ``theta``, or into a list containing a single :class:`~.PauliRot`
            for non-zero ``theta``.

        .. note::

            This operation is used in a differentiation pipeline of :class:`~.SpecialUnitary`
            and most likely should not be created manually by users.
        """
        if qml.math.isclose(theta, theta * 0):
            return []
        return [PauliRot(theta, pauli_word, wires)]

    def __repr__(self):
        return f"TmpPauliRot({self.data[0]}, {self.hyperparameters['pauli_word']}, wires={self.wires.tolist()})"



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains the Shots class to hold shot-related information."""
# pylint:disable=inconsistent-return-statements
from collections import namedtuple
from typing import Sequence, Tuple

ShotCopies = namedtuple("ShotCopies", ["shots", "copies"])
"""A namedtuple that represents a shot quantity being repeated some number of times.

For example, ``ShotCopies(shots=10, copies=2)`` indicates two executions with 10 shots each for 20 shots total.
"""


def valid_int(s):
    """Returns True if s is a positive integer."""
    return isinstance(s, int) and s > 0


def valid_tuple(s):
    """Returns True if s is a tuple of the form (shots, copies)."""
    return isinstance(s, tuple) and len(s) == 2 and valid_int(s[0]) and valid_int(s[1])


class Shots:
    """
    A data class that stores shot information.

    Args:
        shots (Union[None, int, Sequence[int, Tuple[int, int]]]): Raw shot information

    Defining shots enables users to specify circuit executions, and the Shots class standardizes
    the internal representation of shots. There are three ways to specify shot values:

    * The value ``None``
    * A positive integer
    * A sequence consisting of either positive integers or a tuple-pair of positive integers of the form ``(shots, copies)``

    The tuple-pair of the form ``(shots, copies)`` is represented internally by a namedtuple called
    :class:`~ShotCopies`. The first value is the number of shots to execute, and the second value is the
    number of times to repeat a circuit with that number of shots.

    The ``Shots`` class exposes two properties:

    * ``total_shots``, the total number of shots to be executed
    * ``shot_vector``, the tuple of :class:`~ShotCopies` to be executed

    Instances of this class are static. If an instance is passed to the constructor, that same
    instance is returned. If an instance is constructed with a ``None`` value, ``total_shots``
    will be ``None``.  This indicates analytic execution.

    **Examples**

    Example constructing a Shots instance with ``None``:

    >>> shots = Shots(None)
    >>> shots.total_shots, shots.shot_vector
    (None, ())

    Example constructing a Shots instance with an int:

    >>> shots = Shots(100)
    >>> shots.total_shots, shots.shot_vector
    (100, (ShotCopies(shots=100, copies=1),))

    Example constructing a Shots instance with another instance:

    >>> shots = Shots(100)
    >>> Shots(shots) is shots
    True

    Example constructing a Shots instance with a sequence of ints:

    >>> shots = Shots([100, 200])
    >>> shots.total_shots, shots.shot_vector
    (300, (ShotCopies(shots=100, copies=1), ShotCopies(shots=200, copies=1)))

    Example constructing a Shots instance with a sequence of tuple-pairs:

    >>> shots = Shots(((100, 3), (200, 4),))
    >>> shots.total_shots, shots.shot_vector
    (1100, (ShotCopies(shots=100, copies=3), ShotCopies(shots=200, copies=4)))

    Example constructing a Shots instance with a sequence of both ints and tuple-pairs.
    Note that the first stand-alone ``100`` gets absorbed into the subsequent tuple because the
    shot value matches:

    >>> shots = Shots((10, 100, (100, 3), (200, 4),))
    >>> shots.total_shots, shots.shot_vector
    (1210, (ShotCopies(shots=10, copies=1), ShotCopies(shots=100, copies=4), ShotCopies(shots=200, copies=4)))

    One should also note that specifying a single tuple of length 2 is considered two different
    shot values, and *not* a tuple-pair representing shots and copies to avoid special behaviour
    depending on the iterable type:

    >>> shots = Shots((100, 2))
    >>> shots.total_shots, shots.shot_vector
    (102, (ShotCopies(shots=100, copies=1), ShotCopies(shots=2, copies=1)))

    >>> shots = Shots(((100, 2),))
    >>> shots.total_shots, shots.shot_vector
    (200, (ShotCopies(shots=100, copies=2),))
    """

    total_shots: int = None
    """The total number of shots to be executed."""

    shot_vector: Tuple[ShotCopies] = None
    """The tuple of :class:`~ShotCopies` to be executed. Each element is of the form ``(shots, copies)``."""

    _SHOT_ERROR = ValueError(
        "Shots must be a single positive integer, a tuple pair of the form (shots, copies), or a sequence of these types."
    )

    _frozen = False

    def __new__(cls, shots):
        return shots if isinstance(shots, cls) else object.__new__(cls)

    def __setattr__(self, name, value):
        if self._frozen:
            raise AttributeError(
                "Shots is an immutable class. Consider creating a new instance if you need different shot values."
            )
        return super().__setattr__(name, value)

    def __init__(self, shots):
        if shots is None:
            self.total_shots = None
            self.shot_vector = ()
        elif isinstance(shots, int):
            if shots < 1:
                raise self._SHOT_ERROR
            self.total_shots = shots
            self.shot_vector = (ShotCopies(shots, 1),)
        elif isinstance(shots, Sequence):
            if not all(valid_int(s) or valid_tuple(s) for s in shots):
                raise self._SHOT_ERROR
            self.__all_tuple_init__([s if isinstance(s, tuple) else (s, 1) for s in shots])
        elif isinstance(shots, self.__class__):
            return  # self already _is_ shots as defined by __new__
        else:
            raise self._SHOT_ERROR

        self._frozen = True

    def __all_tuple_init__(self, shots: Sequence[Tuple]):
        res = []
        total_shots = 0
        current_shots, current_copies = shots[0]
        for s in shots[1:]:
            if s[0] == current_shots:
                current_copies += s[1]
            else:
                res.append(ShotCopies(current_shots, current_copies))
                total_shots += current_shots * current_copies
                current_shots, current_copies = s
        self.shot_vector = tuple(res + [ShotCopies(current_shots, current_copies)])
        self.total_shots = total_shots + current_shots * current_copies

    @property
    def has_partitioned_shots(self):
        """
        Evaluates to True if this instance represents either multiple shot
        quantities, or the same shot quantity repeated multiple times.

        Returns:
            bool: whether or not shots are partitioned
        """
        if self.total_shots is None:
            return False
        return len(self.shot_vector) > 1 or self.shot_vector[0].copies > 1



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the ``typing.py`` file."""
import numpy as np
import pytest

import pennylane.numpy as pnp
from pennylane.typing import TensorLike


class TestTensorLike:
    def test_isinstance_unknown_type(self):
        """Test that an unknown type returns False."""

        # pylint: disable=too-few-public-methods
        class UnknownType:
            """Unknown type."""

        assert not isinstance(UnknownType(), TensorLike)

    def test_isinstance_numpy_array(self):
        """Tests that a numpy array is a Tensor"""
        assert isinstance(np.array(1), TensorLike)

    def test_isinstance_pennylane_tensor(self):
        """Tests that a PennyLane numpy tensor is a Tensor"""
        assert isinstance(pnp.array(1), TensorLike)

    @pytest.mark.jax
    def test_isinstance_jax_array_is_tensor_like(self):
        """Tests that a jax DeviceArray is a Tensor"""
        import jax

        tensor = jax.numpy.array(1)
        assert isinstance(tensor, TensorLike)

    @pytest.mark.torch
    def test_isinstance_torch_tensor_is_tensor_like(self):
        """Tests that a torch Tensor is a Tensor"""
        import torch

        tensor = torch.Tensor(1)
        assert isinstance(tensor, TensorLike)

    @pytest.mark.tf
    def test_isinstance_tf_tensor_is_tensor_like(self):
        """Tests that a tensorflow Tensor is a Tensor"""
        import tensorflow as tf

        tensor = tf.constant([1, 2, 3])
        assert isinstance(tensor, tf.Tensor)
        assert isinstance(tensor, TensorLike)
        var = tf.Variable(9)
        assert isinstance(var, TensorLike)

    def test_subclass_unknown_type(self):
        """Test that an unknown type returns False."""

        # pylint: disable=too-few-public-methods
        class UnknownType:
            """Unknown type."""

        assert not issubclass(UnknownType, TensorLike)

    def test_subclass_numpy_array(self):
        """Tests that a numpy array is a Tensor"""
        assert issubclass(np.ndarray, TensorLike)

    def test_subclass_pennylane_tensor(self):
        """Tests that a PennyLane numpy tensor is a Tensor"""
        assert issubclass(pnp.ndarray, TensorLike)

    @pytest.mark.jax
    def test_subclass_jax_array_is_tensor_like(self):
        """Tests that a jax DeviceArray is a Tensor"""
        import jax

        assert issubclass(jax.numpy.ndarray, TensorLike)

    @pytest.mark.torch
    def test_subclass_torch_tensor_is_tensor_like(self):
        """Tests that a torch Tensor is a Tensor"""
        import torch

        assert issubclass(torch.Tensor, TensorLike)

    @pytest.mark.tf
    def test_subclass_tf_tensor_is_tensor_like(self):
        """Tests that a tensorflow Tensor is a Tensor"""
        import tensorflow as tf

        assert issubclass(tf.Tensor, TensorLike)
        assert issubclass(tf.Variable, TensorLike)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains preprocessings steps that may be called internally
during execution.
"""
import copy

import pennylane as qml
from pennylane import math
from pennylane.tape import QuantumScript


def _convert_op_to_numpy_data(op: qml.operation.Operator) -> qml.operation.Operator:
    if math.get_interface(*op.data) == "numpy":
        return op
    # Use operator method to change parameters when it become available
    copied_op = copy.copy(op)
    copied_op.data = math.unwrap(op.data)
    return copied_op


def _convert_measurement_to_numpy_data(
    m: qml.measurements.MeasurementProcess,
) -> qml.measurements.MeasurementProcess:
    if m.obs is None or math.get_interface(*m.obs.data) == "numpy":
        return m
    # Use measurement method to change parameters when it becomes available
    copied_m = copy.copy(m)
    if isinstance(copied_m.obs, qml.operation.Tensor):
        copied_m.obs.data = math.unwrap([o.data for o in m.obs.obs])
    else:
        copied_m.obs.data = math.unwrap(m.obs.data)
    return copied_m


# pylint: disable=protected-access
def convert_to_numpy_parameters(circuit: QuantumScript) -> QuantumScript:
    """Transforms a circuit to one with purely numpy parameters.

    Args:
        circuit (QuantumScript): a circuit with parameters of any interface

    Returns:
        QuantumScript: A circuit with purely numpy parameters

    .. seealso::

        :class:`pennylane.tape.Unwrap` modifies a :class:`~.pennylane.tape.QuantumScript` in place instead of creating
        a new class. It will also set all parameters on the circuit, not just ones that need to be unwrapped.

    >>> ops = [qml.S(0), qml.RX(torch.tensor(0.1234), 0)]
    >>> measurements = [qml.state(), qml.expval(qml.Hermitian(torch.eye(2), 0))]
    >>> circuit = qml.tape.QuantumScript(ops, measurements )
    >>> new_circuit = convert_to_numpy_parameters(circuit)
    >>> new_circuit.circuit
    [S(wires=[0]),
    RX(0.1234000027179718, wires=[0]),
    state(wires=[]),
    expval(Hermitian(array([[1., 0.],
            [0., 1.]], dtype=float32), wires=[0]))]

    If the component's data does not need to be transformed, it is left uncopied.

    >>> circuit[0] is new_circuit[0]
    True
    >>> circuit[1] is new_circuit[1]
    False
    >>> circuit[2] is new_circuit[2]
    True
    >>> circuit[3] is new_circuit[3]
    False

    """
    new_prep = (_convert_op_to_numpy_data(op) for op in circuit._prep)
    new_ops = (_convert_op_to_numpy_data(op) for op in circuit._ops)
    new_measurements = (_convert_measurement_to_numpy_data(m) for m in circuit.measurements)
    new_circuit = circuit.__class__(new_ops, new_measurements, new_prep)
    # must preserve trainable params as we lose information about the machine learning interface
    new_circuit.trainable_params = circuit.trainable_params
    new_circuit._qfunc_output = circuit._qfunc_output
    return new_circuit



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains functions for computing the Hadamard-test gradient
of a qubit-based quantum tape.
"""
import pennylane as qml
import pennylane.numpy as np
from pennylane.measurements import MutualInfoMP, StateMP, VarianceMP, VnEntropyMP
from pennylane.transforms.metric_tensor import _get_aux_wire
from pennylane.transforms.tape_expand import expand_invalid_trainable_hadamard_gradient
from .finite_difference import _all_zero_grad_new, _no_trainable_grad_new

from .gradient_transform import (
    choose_grad_methods,
    grad_method_validation,
    gradient_analysis,
    gradient_transform,
)


def _hadamard_grad(
    tape,
    argnum=None,
    shots=None,
    aux_wire=None,
    device_wires=None,
):
    r"""Transform a QNode to compute the Hadamard test gradient of all gates with respect to their inputs.

    Args:
        tape (pennylane.QNode or .QuantumTape): quantum tape or QNode to differentiate
        argnum (int or list[int] or None): Trainable tape parameter indices to differentiate
            with respect to. If not provided, the derivatives with respect to all
            trainable parameters are returned.
        shots (None, int, list[int]): The device shots that will be used to execute the tapes outputted by this
            transform. Note that this argument doesn't influence the shots used for tape execution, but provides
            information about the shots.
        aux_wire (pennylane.wires.Wires): Auxiliary wire to be used for the Hadamard tests. If ``None`` (the default),
            a suitable wire is inferred from the wires used in the original circuit and ``device_wires``.
        device_wires (pennylane.wires.Wires): Wires of the device that are going to be used for the
            gradient. Facilitates finding a default for ``aux_wire`` if ``aux_wire``
            is ``None``.

    Returns:
        function or tuple[list[QuantumTape], function]:

        - If the input is a QNode, an object representing the Jacobian (function) of the QNode
          that can be executed to obtain the Jacobian.
          The type of the Jacobian returned is either a tensor, a tuple or a
          nested tuple depending on the nesting structure of the original QNode output.

        - If the input is a tape, a tuple containing a
          list of generated tapes, together with a post-processing
          function to be applied to the results of the evaluated tapes
          in order to obtain the Jacobian.

    For a variational evolution :math:`U(\mathbf{p}) \vert 0\rangle` with
    :math:`N` parameters :math:`\mathbf{p}`,
    consider the expectation value of an observable :math:`O`:

    .. math::

        f(\mathbf{p})  = \langle \hat{O} \rangle(\mathbf{p}) = \langle 0 \vert
        U(\mathbf{p})^\dagger \hat{O} U(\mathbf{p}) \vert 0\rangle.


    The gradient of this expectation value can be calculated via the Hadamard test gradient:

    .. math::

        \frac{\partial f}{\partial \mathbf{p}} = -2 \Im[\bra{0} \hat{O} G \ket{0}] = i \left(\bra{0} \hat{O} G \ket{
        0} - \bra{0} G\hat{O} \ket{0}\right) = -2 \bra{+}\bra{0} ctrl-G^{\dagger} (\hat{Y} \otimes \hat{O}) ctrl-G
        \ket{+}\ket{0}

    Here, :math:`G` is the generator of the unitary :math:`U`.

    **Example**

    This gradient transform can be applied directly to :class:`QNode <pennylane.QNode>` objects:

    >>> qml.enable_return()
    >>> dev = qml.device("default.qubit", wires=2)
    >>> @qml.qnode(dev)
    ... def circuit(params):
    ...     qml.RX(params[0], wires=0)
    ...     qml.RY(params[1], wires=0)
    ...     qml.RX(params[2], wires=0)
    ...     return qml.expval(qml.PauliZ(0))
    >>> params = np.array([0.1, 0.2, 0.3], requires_grad=True)
    >>> qml.gradients.hadamard_grad(circuit)(params)
    (tensor([-0.3875172], requires_grad=True),
     tensor([-0.18884787], requires_grad=True),
     tensor([-0.38355704], requires_grad=True))

    This quantum gradient transform can also be applied to low-level
    :class:`~.QuantumTape` objects. This will result in no implicit quantum
    device evaluation. Instead, the processed tapes, and post-processing
    function, which together define the gradient are directly returned:

    >>> with qml.tape.QuantumTape() as tape:
    ...     qml.RX(params[0], wires=0)
    ...     qml.RY(params[1], wires=0)
    ...     qml.RX(params[2], wires=0)
    ...     qml.expval(qml.PauliZ(0))
    >>> gradient_tapes, fn = qml.gradients.hadamard_grad(tape)
    >>> gradient_tapes
    [<QuantumTape: wires=[0, 1], params=3>,
     <QuantumTape: wires=[0, 1], params=3>,
     <QuantumTape: wires=[0, 1], params=3>]

    This can be useful if the underlying circuits representing the gradient
    computation need to be analyzed.

    The output tapes can then be evaluated and post-processed to retrieve
    the gradient:

    >>> dev = qml.device("default.qubit", wires=2)
    >>> fn(qml.execute(gradient_tapes, dev, None))
    (array(-0.3875172), array(-0.18884787), array(-0.38355704))

    This transform can be registered directly as the quantum gradient transform
    to use during autodifferentiation:

    >>> dev = qml.device("default.qubit", wires=3)
    >>> @qml.qnode(dev, interface="jax", diff_method="hadamard")
    ... def circuit(params):
    ...     qml.RX(params[0], wires=0)
    ...     qml.RY(params[1], wires=0)
    ...     qml.RX(params[2], wires=0)
    ...     return qml.expval(qml.PauliZ(0))
    >>> params = jax.numpy.array([0.1, 0.2, 0.3])
    >>> jax.jacobian(circuit)(params)
    [-0.3875172  -0.18884787 -0.38355704]

    If you use custom wires on your device, you need to pass an auxiliary wire.

    >>> qml.enable_return()
    >>> dev_wires = ("a", "c")
    >>> dev = qml.device("default.qubit", wires=dev_wires)
    >>> @qml.qnode(dev, interface="jax", diff_method="hadamard", aux_wire="c", device_wires=dev_wires)
    >>> def circuit(params):
    ...    qml.RX(params[0], wires="a")
    ...    qml.RY(params[1], wires="a")
    ...    qml.RX(params[2], wires="a")
    ...    return qml.expval(qml.PauliZ("a"))
    >>> params = jax.numpy.array([0.1, 0.2, 0.3])
    >>> jax.jacobian(circuit)(params)
    [-0.3875172  -0.18884787 -0.38355704]

    .. note::

        ``hadamard_grad`` will decompose the operations that are not in the list of supported operations.

        - ``pennylane.RX``
        - ``pennylane.RY``
        - ``pennylane.RZ``
        - ``pennylane.Rot``
        - ``pennylane.PhaseShift``
        - ``pennylane.U1``
        - ``pennylane.CRX``
        - ``pennylane.CRY``
        - ``pennylane.CRZ``
        - ``pennylane.IsingXX``
        - ``pennylane.IsingYY``
        - ``pennylane.IsingZZ``

        The expansion will fail if a suitable decomposition in terms of supported operation is not found.
        The number of trainable parameters may increase due to the decomposition.

    """
    if not qml.active_return():
        raise NotImplementedError(
            "The Hadamard gradient only supports the new return type. Use qml.enable_return() to turn it on."
        )
    if any(isinstance(m, VarianceMP) for m in tape.measurements):
        raise ValueError(
            "Computing the gradient of variances with the Hadamard test gradient is not implemented."
        )
    if any(isinstance(m, (StateMP, VnEntropyMP, MutualInfoMP)) for m in tape.measurements):
        raise ValueError(
            "Computing the gradient of circuits that return the state is not supported."
        )

    if argnum is None and not tape.trainable_params:
        return _no_trainable_grad_new(tape, shots)

    gradient_analysis(tape, grad_fn=hadamard_grad)
    method = "analytic"
    diff_methods = grad_method_validation(method, tape)

    if all(g == "0" for g in diff_methods):
        return _all_zero_grad_new(tape, shots)

    method_map = choose_grad_methods(diff_methods, argnum)

    argnum = [i for i, dm in method_map.items() if dm == "A"]

    if device_wires and len(tape.wires) == len(device_wires):
        raise qml.QuantumFunctionError("The device has no free wire for the auxiliary wire.")

    # Get default for aux_wire
    if aux_wire is None:
        aux_wire = _get_aux_wire(aux_wire, tape, device_wires)
    elif aux_wire[0] in tape.wires:
        raise qml.QuantumFunctionError("The auxiliary wire is already used.")
    elif aux_wire[0] not in device_wires:
        raise qml.QuantumFunctionError(
            "The requested auxiliary wire does not exist on the used device."
        )

    g_tapes, processing_fn = _expval_hadamard_grad(tape, argnum, aux_wire)

    return g_tapes, processing_fn


def _expval_hadamard_grad(tape, argnum, aux_wire):
    r"""Compute the Hadamard test gradient of a tape that returns an expectation value (probabilities are expectations
    values) with respect to a given set of all trainable gate parameters.
    The auxiliary wire is the wire which is used to apply the Hadamard gates and controlled gates.
    """
    # pylint: disable=too-many-statements
    argnums = argnum or tape.trainable_params
    g_tapes = []
    coeffs = []

    gradient_data = []
    measurements_probs = [
        idx
        for idx, m in enumerate(tape.measurements)
        if isinstance(m, qml.measurements.ProbabilityMP)
    ]
    for id_argnum, _ in enumerate(tape.trainable_params):
        if id_argnum not in argnums:
            # parameter has zero gradient
            gradient_data.append(0)
            continue

        trainable_op, idx, p_idx = tape.get_operation(id_argnum)

        ops_to_trainable_op = tape.operations[: idx + 1]
        ops_after_trainable_op = tape.operations[idx + 1 :]

        # Get a generator and coefficients
        sub_coeffs, generators = _get_generators(trainable_op)
        coeffs.extend(sub_coeffs)

        num_tape = 0

        for gen in generators:
            if isinstance(trainable_op, qml.Rot):
                # We only registered PauliZ as generator for Rot, therefore we need to apply some gates
                # before and after the generator for the first two parameters.
                if p_idx == 0:
                    # Move the Rot gate past the generator
                    op_before_trainable_op = ops_to_trainable_op.pop(-1)
                    ops_after_trainable_op = [op_before_trainable_op] + ops_after_trainable_op
                elif p_idx == 1:
                    # Apply additional rotations that effectively move the generator to the middle of Rot
                    ops_to_add_before = [
                        qml.RZ(-trainable_op.data[2], wires=trainable_op.wires),
                        qml.RX(np.pi / 2, wires=trainable_op.wires),
                    ]
                    ops_to_trainable_op.extend(ops_to_add_before)

                    ops_to_add_after = [
                        qml.RX(-np.pi / 2, wires=trainable_op.wires),
                        qml.RZ(trainable_op.data[2], wires=trainable_op.wires),
                    ]
                    ops_after_trainable_op = ops_to_add_after + ops_after_trainable_op

            ctrl_gen = [qml.ctrl(gen, control=aux_wire)]
            hadamard = [qml.Hadamard(wires=aux_wire)]
            ops = ops_to_trainable_op + hadamard + ctrl_gen + hadamard + ops_after_trainable_op

            measurements = []
            # Add the Y measurement on the aux qubit
            for m in tape.measurements:
                if isinstance(m.obs, qml.operation.Tensor):
                    obs_new = m.obs.obs.copy()
                elif m.obs:
                    obs_new = [m.obs]
                else:
                    obs_new = [qml.PauliZ(wires=i) for i in m.wires]

                obs_new.append(qml.PauliY(wires=aux_wire))
                obs_new = qml.operation.Tensor(*obs_new)

                if isinstance(m, qml.measurements.ExpectationMP):
                    measurements.append(qml.expval(op=obs_new))
                else:
                    measurements.append(qml.probs(op=obs_new))

            new_tape = qml.tape.QuantumScript(ops=ops, measurements=measurements)

            _rotations, _measurements = qml.tape.tape.rotations_and_diagonal_measurements(new_tape)
            # pylint: disable=protected-access
            new_tape._ops = new_tape._ops + _rotations
            new_tape._measurements = _measurements
            new_tape._update()

            num_tape += 1

            g_tapes.append(new_tape)

        gradient_data.append(num_tape)

    def processing_fn(results):  # pylint: disable=too-many-branches
        multi_measurements = len(tape.measurements) > 1
        multi_params = len(tape.trainable_params) > 1

        final_res = [
            [qml.math.convert_like(2 * coeff * r, r) for r in res]
            if isinstance(res, tuple)
            else qml.math.convert_like(2 * coeff * res, res)
            for coeff, res in zip(coeffs, results)
        ]

        # Post process for probs
        if measurements_probs:
            projector = np.array([1, -1])
            if multi_measurements:
                projector = qml.math.convert_like(projector, final_res[0][0])
            else:
                projector = qml.math.convert_like(projector, final_res[0])
            for idx, res in enumerate(final_res):
                if multi_measurements:
                    for prob_idx in measurements_probs:
                        num_wires_probs = len(tape.measurements[prob_idx].wires)
                        res_reshaped = qml.math.reshape(res[prob_idx], (2**num_wires_probs, 2))
                        final_res[idx][prob_idx] = qml.math.tensordot(
                            res_reshaped, projector, axes=[[1], [0]]
                        )
                else:
                    prob_idx = measurements_probs[0]
                    num_wires_probs = len(tape.measurements[prob_idx].wires)
                    res = qml.math.reshape(res, (2**num_wires_probs, 2))
                    final_res[idx] = qml.math.tensordot(res, projector, axes=[[1], [0]])
        grads = []
        idx = 0
        for num_tape in gradient_data:
            if num_tape == 0:
                grads.append(qml.math.zeros(()))
            elif num_tape == 1:
                grads.append(final_res[idx])
                idx += 1
            else:
                axis = None if not multi_measurements else 0
                grads.append(qml.math.sum(final_res[idx : idx + num_tape], axis=axis))
                idx += num_tape

        if not multi_measurements and not multi_params:
            return grads[0]

        if not (multi_params and multi_measurements):
            if multi_measurements:
                return tuple(grads[0])
            return tuple(grads)

        # Reordering to match the right shape for multiple measurements
        grads_reorder = [[0] * len(tape.trainable_params) for _ in range(len(tape.measurements))]

        for i in range(len(tape.measurements)):
            for j in range(len(tape.trainable_params)):
                grads_reorder[i][j] = grads[j][i]

        grads_tuple = tuple(tuple(elem) for elem in grads_reorder)

        return grads_tuple

    return g_tapes, processing_fn


def _get_generators(trainable_op):
    """From a trainable operation, extract the unitary generators and their coefficients. If an operation is added here
    one needs to also update the list of supported operation in the expand function given to the gradient transform.
    """
    # For PhaseShift, we need to separate the generator in two unitaries (Hardware compatibility)
    if isinstance(trainable_op, (qml.PhaseShift, qml.U1)):
        generators = [qml.PauliZ(wires=trainable_op.wires)]
        coeffs = [-0.5]
    elif isinstance(trainable_op, qml.CRX):
        generators = [
            qml.PauliX(wires=trainable_op.wires[1]),
            qml.prod(
                qml.PauliZ(wires=trainable_op.wires[0]), qml.PauliX(wires=trainable_op.wires[1])
            ),
        ]
        coeffs = [-0.25, 0.25]
    elif isinstance(trainable_op, qml.CRY):
        generators = [
            qml.PauliY(wires=trainable_op.wires[1]),
            qml.prod(
                qml.PauliZ(wires=trainable_op.wires[0]), qml.PauliY(wires=trainable_op.wires[1])
            ),
        ]
        coeffs = [-0.25, 0.25]
    elif isinstance(trainable_op, qml.CRZ):
        generators = [
            qml.PauliZ(wires=trainable_op.wires[1]),
            qml.prod(
                qml.PauliZ(wires=trainable_op.wires[0]), qml.PauliZ(wires=trainable_op.wires[1])
            ),
        ]
        coeffs = [-0.25, 0.25]
    elif isinstance(trainable_op, qml.IsingXX):
        generators = [
            qml.prod(
                qml.PauliX(wires=trainable_op.wires[0]), qml.PauliX(wires=trainable_op.wires[1])
            )
        ]
        coeffs = [-0.5]
    elif isinstance(trainable_op, qml.IsingYY):
        generators = [
            qml.prod(
                qml.PauliY(wires=trainable_op.wires[0]), qml.PauliY(wires=trainable_op.wires[1])
            )
        ]
        coeffs = [-0.5]
    elif isinstance(trainable_op, qml.IsingZZ):
        generators = [
            qml.prod(
                qml.PauliZ(wires=trainable_op.wires[0]), qml.PauliZ(wires=trainable_op.wires[1])
            )
        ]
        coeffs = [-0.5]
    # For rotation it is possible to only use PauliZ by applying some other rotations in the main function
    elif isinstance(trainable_op, qml.Rot):
        generators = [qml.PauliZ(wires=trainable_op.wires)]
        coeffs = [-0.5]
    else:
        generators = trainable_op.generator().ops
        coeffs = trainable_op.generator().coeffs

    return coeffs, generators


hadamard_grad = gradient_transform(
    _hadamard_grad, expand_fn=expand_invalid_trainable_hadamard_gradient
)



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the QSVT template and qsvt wrapper function.
"""
# pylint: disable=too-many-arguments
import copy
import pennylane as qml
from pennylane.queuing import QueuingManager
from pennylane.ops import BlockEncode, PCPhase
from pennylane.ops.op_math import adjoint
from pennylane.operation import AnyWires, Operation
from pennylane import numpy as np


def qsvt(A, angles, wires, convention=None):
    r"""Implements the
    `quantum singular value transformation <https://arxiv.org/abs/1806.01838>`__ (QSVT) circuit.

    .. note ::

        :class:`~.BlockEncode` and :class:`~.PCPhase` used in this implementation of QSVT
        are matrix-based operators and well-suited for simulators.
        To implement QSVT with user-defined circuits for the block encoding and
        projector-controlled phase shifts, use the :class:`~.QSVT` template.

    Given a matrix :math:`A`, and a list of angles :math:`\vec{\phi}`, this function applies a
    circuit for the quantum singular value transformation using :class:`~.BlockEncode` and
    :class:`~.PCPhase`.

    When the number of angles is even (:math:`d` is odd), the QSVT circuit is defined as:

    .. math::

        U_{QSVT} = \tilde{\Pi}_{\phi_1}U\left[\prod^{(d-1)/2}_{k=1}\Pi_{\phi_{2k}}U^\dagger
        \tilde{\Pi}_{\phi_{2k+1}}U\right]\Pi_{\phi_{d+1}},


    and when the number of angles is odd (:math:`d` is even):

    .. math::

        U_{QSVT} = \left[\prod^{d/2}_{k=1}\Pi_{\phi_{2k-1}}U^\dagger\tilde{\Pi}_{\phi_{2k}}U\right]
        \Pi_{\phi_{d+1}}.

    Here, :math:`U` denotes a block encoding of :math:`A` via :class:`~.BlockEncode` and
    :math:`\Pi_\phi` denotes a projector-controlled phase shift with angle :math:`\phi`
    via :class:`~.PCPhase`.

    This circuit applies a polynomial transformation (:math:`Poly^{SV}`) to the singular values of
    the block encoded matrix:

    .. math::

        \begin{align}
             U_{QSVT}(A, \phi) &=
             \begin{bmatrix}
                Poly^{SV}(A) & \cdot \\
                \cdot & \cdot
            \end{bmatrix}.
        \end{align}

    The polynomial transformation is determined by a combination of the block encoding and choice of angles,
    :math:`\vec{\phi}`. The convention used by :class:`~.BlockEncode` is commonly refered to as the
    reflection convention or :math:`R` convention. Another equivalent convention for the block encoding is
    the :math:`Wx` or rotation convention.

    Depending on the choice of convention for blockencoding, the same phase angles will produce different
    polynomial transformations. We provide the functionality to swap between blockencoding conventions and
    to transform the phase angles accordingly using the :code:`convention` keyword argument.

    .. seealso::

        :class:`~.QSVT` and `A Grand Unification of Quantum Algorithms <https://arxiv.org/pdf/2105.02859.pdf>`_.

    Args:
        A (tensor_like): the general :math:`(n \times m)` matrix to be encoded
        angles (tensor_like): a list of angles by which to shift to obtain the desired polynomial
        wires (Iterable[int, str], Wires): the wires the template acts on
        convention (string): can be set to ``"Wx"`` to convert quantum signal processing angles in the
            `Wx` convention to QSVT angles.

    **Example**

    To implement QSVT in a circuit, we can use the following method:

    >>> dev = qml.device("default.qubit", wires=2)
    >>> A = np.array([[0.1, 0.2], [0.3, 0.4]])
    >>> angles = np.array([0.1, 0.2, 0.3])
    >>> @qml.qnode(dev)
    ... def example_circuit(A):
    ...     qml.qsvt(A, angles, wires=[0, 1])
    ...     return qml.expval(qml.PauliZ(wires=0))

    The resulting circuit implements QSVT.

    >>> print(qml.draw(example_circuit)(A))
    0: ─╭QSVT─┤  <Z>
    1: ─╰QSVT─┤

    To see the implementation details, we can expand the circuit:

    >>> q_script = qml.tape.QuantumScript(ops=[qml.qsvt(A, angles, wires=[0, 1])])
    >>> print(q_script.expand().draw(decimals=2))
    0: ─╭∏_ϕ(0.30)─╭BlockEncode(M0)─╭∏_ϕ(0.20)─╭BlockEncode(M0)†─╭∏_ϕ(0.10)─┤
    1: ─╰∏_ϕ(0.30)─╰BlockEncode(M0)─╰∏_ϕ(0.20)─╰BlockEncode(M0)†─╰∏_ϕ(0.10)─┤
    """
    if qml.math.shape(A) == () or qml.math.shape(A) == (1,):
        A = qml.math.reshape(A, [1, 1])

    c, r = qml.math.shape(A)

    UA = BlockEncode(A, wires=wires, do_queue=False)
    projectors = []

    if convention == "Wx":
        angles = _qsp_to_qsvt(angles)
        global_phase = (len(angles) - 1) % 4

        if global_phase:
            qml.exp(qml.Identity(wires=wires), 0.5j * np.pi * (4 - global_phase))

    for idx, phi in enumerate(angles):
        dim = c if idx % 2 else r
        projectors.append(PCPhase(phi, dim=dim, wires=wires, do_queue=False))

    projectors = projectors[::-1]  # reverse order to match equation
    return QSVT(UA, projectors)


class QSVT(Operation):
    r"""QSVT(UA,projectors)
    Implements the
    `quantum singular value transformation <https://arxiv.org/abs/1806.01838>`__ (QSVT) circuit.

    .. note ::

        This template allows users to define hardware-compatible block encoding and
        projector-controlled phase shift circuits. For a QSVT implementation that is
        tailored for simulators see :func:`~.qsvt` .

    Given an :class:`~.Operator` :math:`U`, which block encodes the matrix :math:`A`, and a list of
    projector-controlled phase shift operations :math:`\vec{\Pi}_\phi`, this template applies a
    circuit for the quantum singular value transformation as follows.

    When the number of projector-controlled phase shifts is even (:math:`d` is odd), the QSVT
    circuit is defined as:

    .. math::

        U_{QSVT} = \tilde{\Pi}_{\phi_1}U\left[\prod^{(d-1)/2}_{k=1}\Pi_{\phi_{2k}}U^\dagger
        \tilde{\Pi}_{\phi_{2k+1}}U\right]\Pi_{\phi_{d+1}}.


    And when the number of projector-controlled phase shifts is odd (:math:`d` is even):

    .. math::

        U_{QSVT} = \left[\prod^{d/2}_{k=1}\Pi_{\phi_{2k-1}}U^\dagger\tilde{\Pi}_{\phi_{2k}}U\right]
        \Pi_{\phi_{d+1}}.

    This circuit applies a polynomial transformation (:math:`Poly^{SV}`) to the singular values of
    the block encoded matrix:

    .. math::

        \begin{align}
             U_{QSVT}(A, \vec{\phi}) &=
             \begin{bmatrix}
                Poly^{SV}(A) & \cdot \\
                \cdot & \cdot
            \end{bmatrix}.
        \end{align}

    .. seealso::

        :func:`~.qsvt` and `A Grand Unification of Quantum Algorithms <https://arxiv.org/pdf/2105.02859.pdf>`_.

    Args:
        UA (Operator): the block encoding circuit, specified as an :class:`~.Operator`,
            like :class:`~.BlockEncode`
        projectors (Sequence[Operator]): a list of projector-controlled phase
            shifts that implement the desired polynomial

    Raises:
        ValueError: if the input block encoding is not an operator

    **Example**

    To implement QSVT in a circuit, we can use the following method:

    >>> dev = qml.device("default.qubit", wires=2)
    >>> A = np.array([[0.1]])
    >>> block_encode = qml.BlockEncode(A, wires=[0, 1])
    >>> shifts = [qml.PCPhase(i + 0.1, dim=1, wires=[0, 1]) for i in range(3)]
    >>> @qml.qnode(dev)
    >>> def example_circuit():
    ...    qml.QSVT(block_encode, shifts)
    ...    return qml.expval(qml.PauliZ(wires=0))

    The resulting circuit implements QSVT.

    >>> print(qml.draw(example_circuit)())
    0: ─╭QSVT─┤  <Z>
    1: ─╰QSVT─┤

    To see the implementation details, we can expand the circuit:

    >>> q_script = qml.tape.QuantumScript(ops=[qml.QSVT(block_encode, shifts)])
    >>> print(q_script.expand().draw(decimals=2))
    0: ─╭∏_ϕ(0.10)─╭BlockEncode(M0)─╭∏_ϕ(1.10)─╭BlockEncode(M0)†─╭∏_ϕ(2.10)─┤
    1: ─╰∏_ϕ(0.10)─╰BlockEncode(M0)─╰∏_ϕ(1.10)─╰BlockEncode(M0)†─╰∏_ϕ(2.10)─┤

    When working with this class directly, we can make use of any PennyLane operation
    to represent our block-encoding and our phase-shifts.

    >>> dev = qml.device("default.qubit", wires=[0])
    >>> block_encoding = qml.Hadamard(wires=0)  # note H is a block encoding of 1/sqrt(2)
    >>> phase_shifts = [qml.RZ(-2 * theta, wires=0) for theta in (1.23, -0.5, 4)]  # -2*theta to match convention
    >>>
    >>> @qml.qnode(dev)
    >>> def example_circuit():
    ...     qml.QSVT(block_encoding, phase_shifts)
    ...     return qml.expval(qml.PauliZ(wires=0))
    >>>
    >>> example_circuit()
    tensor(0.54030231, requires_grad=True)

    Once again, we can visualize the circuit as follows:

    >>> print(qml.draw(example_circuit)())
    0: ──QSVT─┤  <Z>

    To see the implementation details, we can expand the circuit:

    >>> q_script = qml.tape.QuantumScript(ops=[qml.QSVT(block_encoding, phase_shifts)])
    >>> print(q_script.expand().draw(decimals=2))
    0: ──RZ(-2.46)──H──RZ(1.00)──H†──RZ(-8.00)─┤
    """

    num_wires = AnyWires
    """int: Number of wires that the operator acts on."""

    grad_method = None
    """Gradient computation method."""

    def __init__(self, UA, projectors, do_queue=True, id=None):
        if not isinstance(UA, qml.operation.Operator):
            raise ValueError("Input block encoding must be an Operator")

        self._hyperparameters = {
            "UA": UA,
            "projectors": projectors,
        }

        ua_wires = UA.wires.toset()
        proj_wires = set.union(*(proj.wires.toset() for proj in projectors))

        total_wires = ua_wires.union(proj_wires)
        super().__init__(wires=total_wires, do_queue=do_queue, id=id)

    @staticmethod
    def compute_decomposition(
        UA, projectors, **hyperparameters
    ):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a product of other operators.

        The :class:`~.QSVT` is decomposed into alternating block encoding
        and projector-controlled phase shift operators. This is defined by the following
        equations, where :math:`U` is the block encoding operation and both :math:`\Pi_\phi` and
        :math:`\tilde{\Pi}_\phi` are projector-controlled phase shifts with angle :math:`\phi`.

        When the number of projector-controlled phase shifts is even (:math:`d` is odd), the QSVT
        circuit is defined as:

        .. math::

            U_{QSVT} = \Pi_{\phi_1}U\left[\prod^{(d-1)/2}_{k=1}\Pi_{\phi_{2k}}U^\dagger
            \tilde{\Pi}_{\phi_{2k+1}}U\right]\Pi_{\phi_{d+1}}.


        And when the number of projector-controlled phase shifts is odd (:math:`d` is even):

        .. math::

            U_{QSVT} = \left[\prod^{d/2}_{k=1}\Pi_{\phi_{2k-1}}U^\dagger\tilde{\Pi}_{\phi_{2k}}U\right]
            \Pi_{\phi_{d+1}}.

        .. seealso:: :meth:`~.QSVT.decomposition`.

        Args:
            UA (Operator): the block encoding circuit, specified as a :class:`~.Operator`
            projectors (list[Operator]): a list of projector-controlled phase
                shift circuits that implement the desired polynomial
            wires (Iterable): wires that the template acts on

        Returns:
            list[.Operator]: decomposition of the operator
        """

        op_list = []
        UA_adj = copy.copy(UA)

        for idx, op in enumerate(projectors[:-1]):
            qml.apply(op)
            op_list.append(op)

            if idx % 2 == 0:
                qml.apply(UA)
                op_list.append(UA)

            else:
                op_list.append(adjoint(UA_adj))

        qml.apply(projectors[-1])
        op_list.append(projectors[-1])

        return op_list

    def label(self, decimals=None, base_label=None, cache=None):
        op_label = base_label or self.__class__.__name__
        return op_label

    def queue(self, context=QueuingManager):
        context.remove(self._hyperparameters["UA"])
        for op in self._hyperparameters["projectors"]:
            context.remove(op)
        context.append(self)
        return self

    @staticmethod
    def compute_matrix(*args, **kwargs):
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.Operator.matrix` and :func:`~.matrix`

        Args:
            params (list): trainable parameters of the operator, as stored in the ``parameters`` attribute
            hyperparams (dict): non-trainable hyperparameters of the operator, as stored in the ``hyperparameters`` attribute

        Returns:
            tensor_like: matrix representation
        """
        # pylint: disable=unused-argument
        op_list = []
        UA = kwargs["UA"]
        projectors = kwargs["projectors"]

        with QueuingManager.stop_recording():  # incase this method is called in a queue context, this prevents
            UA_copy = copy.copy(UA)  # us from queuing operators unnecessarily

            for idx, op in enumerate(projectors[:-1]):
                op_list.append(op)
                if idx % 2 == 0:
                    op_list.append(UA)
                else:
                    op_list.append(adjoint(UA_copy))

            op_list.append(projectors[-1])
            mat = qml.matrix(qml.prod(*tuple(op_list[::-1])))

        return mat


def _qsp_to_qsvt(angles):
    r"""Converts qsp angles to qsvt angles."""
    new_angles = qml.math.array(copy.copy(angles))
    new_angles[0] += 3 * np.pi / 4
    new_angles[-1] -= np.pi / 4

    new_angles[1:-1] += np.pi / 2
    return new_angles



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This submodule contains controlled operators based on the Controlled and ControlledOp class.
"""

import warnings
from typing import Iterable

from pennylane.operation import AnyWires
from pennylane.ops.qubit.matrix_ops import QubitUnitary
from .controlled import ControlledOp


# pylint: disable=too-few-public-methods
class ControlledQubitUnitary(ControlledOp):
    r"""ControlledQubitUnitary(U, control_wires, wires, control_values)
    Apply an arbitrary fixed unitary to ``wires`` with control from the ``control_wires``.

    In addition to default ``Operation`` instance attributes, the following are
    available for ``ControlledQubitUnitary``:

    * ``control_wires``: wires that act as control for the operation
    * ``control_values``: the state on which to apply the controlled operation (see below)
    * ``target_wires``: the wires the unitary matrix will be applied to
    * ``active_wires``: Wires modified by the operator. This is the control wires followed by the target wires.

    **Details:**

    * Number of wires: Any (the operation can act on any number of wires)
    * Number of parameters: 1
    * Number of dimensions per parameter: (2,)
    * Gradient recipe: None

    Args:
        base (Union[array[complex], QubitUnitary]): square unitary matrix or a QubitUnitary operation. If passing a matrix,
          this will be used to construct a QubitUnitary operator that will be used as the base operator. If providing
          a ``qml.QubitUnitary``, this will be used as the base directly.
        control_wires (Union[Wires, Sequence[int], or int]): the control wire(s)
        wires (Union[Wires, Sequence[int], or int]): the wire(s) the unitary acts on (optional if U is provided as a QubitUnitary)
        control_values (List[int, bool]): a list providing the state of the control qubits to control on (default is the all 1s state)
        unitary_check (bool): whether to check whether an array U is unitary when creating the operator (default False)
        do_queue (bool): indicates whether the operator should be recorded when created in a tape context

    **Example**

    The following shows how a single-qubit unitary can be applied to wire ``2`` with control on
    both wires ``0`` and ``1``:

    >>> U = np.array([[ 0.94877869,  0.31594146], [-0.31594146,  0.94877869]])
    >>> qml.ControlledQubitUnitary(U, control_wires=[0, 1], wires=2)
    Controlled(QubitUnitary(array([[ 0.94877869,  0.31594146],
       [-0.31594146,  0.94877869]]), wires=[2]), control_wires=[0, 1])

    Alternatively, the same operator can be constructed with a QubitUnitary:

    >>> base = qml.QubitUnitary(U, wires=2)
    >>> qml.ControlledQubitUnitary(base, control_wires=[0, 1])
    Controlled(QubitUnitary(array([[ 0.94877869,  0.31594146],
       [-0.31594146,  0.94877869]]), wires=[2]), control_wires=[0, 1])

    Typically controlled operations apply a desired gate if the control qubits
    are all in the state :math:`\vert 1\rangle`. However, there are some situations where
    it is necessary to apply a gate conditioned on all qubits being in the
    :math:`\vert 0\rangle` state, or a mix of the two.

    The state on which to control can be changed by passing a string of bits to
    `control_values`. For example, if we want to apply a single-qubit unitary to
    wire ``3`` conditioned on three wires where the first is in state ``0``, the
    second is in state ``1``, and the third in state ``1``, we can write:

    >>> qml.ControlledQubitUnitary(U, control_wires=[0, 1, 2], wires=3, control_values=[0, 1, 1])

    or

    >>> qml.ControlledQubitUnitary(U, control_wires=[0, 1, 2], wires=3, control_values=[False, True, True])
    """
    num_wires = AnyWires
    """int: Number of wires that the operator acts on."""

    num_params = 1
    """int: Number of trainable parameters that the operator depends on."""

    ndim_params = (2,)
    """tuple[int]: Number of dimensions per trainable parameter that the operator depends on."""

    grad_method = None
    """Gradient computation method."""

    # pylint: disable= too-many-arguments
    def __init__(
        self,
        base,
        control_wires,
        wires=None,
        control_values=None,
        do_queue=True,
        unitary_check=False,
        work_wires=None,
    ):
        if getattr(base, "wires", False) and wires is not None:
            warnings.warn(
                "base operator already has wires; values specified through wires kwarg will be ignored."
            )

        if isinstance(base, Iterable):
            base = QubitUnitary(base, wires=wires, unitary_check=unitary_check)

        super().__init__(
            base,
            control_wires,
            control_values=control_values,
            work_wires=work_wires,
            do_queue=do_queue,
        )
        self._name = "ControlledQubitUnitary"

    def _controlled(self, wire):
        ctrl_wires = self.control_wires + wire
        values = None if self.control_values is None else self.control_values + [True]
        new_op = ControlledQubitUnitary(
            self.base,
            control_wires=ctrl_wires,
            control_values=values,
            work_wires=self.work_wires,
        )
        return new_op



# Copyright 2018-2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains the classes/functions needed to simulate the evolution of ensembles of
individual (trapped) rydberg atoms under the excitation of several laser fields."""
from dataclasses import dataclass
import numpy as np
import pennylane as qml

from pennylane.pulse import HardwareHamiltonian, HardwarePulse, drive
from pennylane.wires import Wires
from pennylane.pulse.hardware_hamiltonian import _reorder_parameters


def rydberg_interaction(
    register: list, wires=None, interaction_coeff: float = 862690, max_distance: float = np.inf
):
    r"""Returns a :class:`ParametrizedHamiltonian` representing the interaction of an ensemble of
    Rydberg atoms due to the Rydberg blockade

    .. math::

        \sum_{i<j} V_{ij} n_i n_j

    where :math:`n_i` corresponds to the projector on the Rydberg state of the atom :math:`i`, and
    :math:`V_{ij}` is the van der Waals potential:

    .. math::

        V_{ij} = \frac{C_6}{R_{ij}^6}

    where :math:`R_{ij}` is the distance between the atoms :math:`i` and :math:`j`, and :math:`C_6`
    is the Rydberg interaction constant, which defaults to :math:`862690 \times 2 \pi \text{MHz } \mu \text{m}^6`.
    The unit of time for the evolution of this Rydberg interaction term is in :math:`\mu \text{s}`.
    This interaction term can be combined with laser drive terms (:func:`~.rydberg_drive`) to create
    a Hamiltonian describing a driven Rydberg atom system.

    .. seealso::

        :func:`~.rydberg_drive`

    Args:
        register (list): list of coordinates of the Rydberg atoms (in micrometers)
        wires (list): List of wires containing the wire values for all the atoms. This list should
            have the same length as ``register``. If ``None``, each atom's wire value will
            correspond to its index in the ``register`` list.
        interaction_coeff (float): Defaults to :math:`862690 \times 2 \pi \text{MHz } \mu\text{m}^6`.
            The value will be multiplied by :math:`2 \pi` internally to convert to angular frequency,
            such that only the value in standard frequency (i.e., 862690 in the default example) should be
            passed.
        max_distance (float): Threshold for distance in :math:`\mu\text{m}` between two Rydberg atoms beyond which their
            contribution to the interaction term is removed from the Hamiltonian.

    Returns:
        :class:`~.ParametrizedHamiltonian`: a :class:`~.ParametrizedHamiltonian` representing the atom interaction

    **Example**

    We create a Hamiltonian describing the van der Waals interaction among 9 Rydberg atoms in a square lattice:

    .. code-block:: python

        atom_coordinates = [[0, 0], [0, 5], [0, 10], [5, 0], [5, 5], [5, 10], [10, 0], [10, 5], [10, 10]]
        wires = [1, 5, 0, 2, 4, 3, 8, 6, 7]
        H_i = qml.pulse.rydberg_interaction(atom_coordinates, wires=wires)

    >>> H_i
    ParametrizedHamiltonian: terms=36

    As expected, we have :math:`\frac{N(N-1)}{2} = 36` terms for N=9 atoms.

    The interaction term is dependent only on the number and positions of the Rydberg atoms. We can execute this
    pulse program, which corresponds to all driving laser fields being turned off and therefore has no trainable
    parameters. To add a driving laser field, see :func:`~.rydberg_drive`.

    .. code-block:: python

        dev = qml.device("default.qubit.jax", wires=9)

        @qml.qnode(dev, interface="jax")
        def circuit():
            qml.evolve(H_i)([], t=[0, 10])
            return qml.expval(qml.PauliZ(0))

    >>> circuit()
    Array(1., dtype=float32)
    """
    if wires is None:
        wires = list(range(len(register)))
    elif len(wires) != len(register):
        raise ValueError("The length of the wires and the register must match.")

    coeffs = []
    observables = []
    for idx, (pos1, wire1) in enumerate(zip(register[:-1], wires[:-1])):
        for pos2, wire2 in zip(register[(idx + 1) :], wires[(idx + 1) :]):
            atom_distance = np.linalg.norm(qml.math.array(pos2) - pos1)
            if atom_distance > max_distance:
                continue
            # factor 2pi converts interaction coefficient from standard to angular frequency
            Vij = (
                2 * np.pi * interaction_coeff / (abs(atom_distance) ** 6)
            )  # van der Waals potential
            coeffs.append(Vij)
            observables.append(qml.prod(qml.Projector([1], wire1), qml.Projector([1], wire2)))
            # Rydberg projectors

    settings = RydbergSettings(register, interaction_coeff)

    return HardwareHamiltonian(
        coeffs, observables, settings=settings, reorder_fn=_reorder_parameters
    )


def rydberg_drive(amplitude, phase, detuning, wires):
    r"""Returns a :class:`ParametrizedHamiltonian` representing the action of a driving laser field

    .. math::

        \frac{1}{2} \Omega(t) \sum_{q \in \text{wires}} (\cos(\phi(t))\sigma_q^x - \sin(\phi(t))\sigma_q^y) -
        \delta(t) \sum_{q \in \text{wires}} n_q

    where :math:`\Omega/(2\pi)`, :math:`\phi` and :math:`\delta/(2\pi)` correspond to the amplitude, phase,
    and detuning of the laser, :math:`q` corresponds to the wire index, and
    :math:`\sigma_q^\alpha` for :math:`\alpha = x,y` are the Pauli matrices on the corresponding
    qubit. Finally, :math:`n_q=\frac{1}{2}(\mathbb{I}_q-\sigma_q^z)` is the number operator on qubit :math:`q`.

    .. note::
        For hardware execution, input time is expected to be in units of :math:`\mu\text{s}`, and the frequency
        in units of MHz. It is recommended to also follow this convention for simulation,
        as it avoids numerical problems due to using very large and very small numbers. Frequency inputs will be
        converted internally to angular frequency, such that ``amplitude`` :math:`= \Omega(t)/ (2 \pi)` and
        ``detuning`` :math:`= \delta(t) / (2 \pi)`.

    This driving term can be combined with an interaction term to create a Hamiltonian describing a
    driven Rydberg atom system. Multiple driving terms can be combined by summing them (see example).

    Args:
        amplitude (Union[float, Callable]): Float or callable representing the amplitude of a laser field.
            This should be in units of frequency (MHz), and will be converted to amplitude in angular frequency,
            :math:`\Omega(t)`, internally where needed, i.e. multiplied by :math:`2 \pi`.
        phase (Union[float, Callable]): float or callable representing the phase (in radians) of the laser field
        detuning (Union[float, Callable]): Float or callable representing the detuning of a laser field.
            This should be in units of frequency (MHz), and will be converted to detuning in angular frequency,
            :math:`\delta(t)`, internally where needed, i.e. multiplied by :math:`2 \pi`.
        wires (Union[int, List[int]]): integer or list containing wire values for the Rydberg atoms that
            the laser field acts on

    Returns:
        :class:`~.ParametrizedHamiltonian`: a :class:`~.ParametrizedHamiltonian` representing the action of the laser field on the Rydberg atoms.

    .. seealso::

        :func:`~.rydberg_interaction`, :class:`~.ParametrizedHamiltonian`, :class:`~.ParametrizedEvolution`
        and :func:`~.evolve`

    **Example**

    We create a Hamiltonian describing a laser acting on 4 wires (Rydberg atoms) with a fixed detuning and
    phase, and a parametrized, time-dependent amplitude. The Hamiltonian includes an interaction term for
    inter-atom interactions due to van der Waals forces, as well as the driving term for the laser driving
    the atoms.

    We provide all frequencies in the driving term in MHz (conversion to angular frequency, i.e. multiplication
    by :math:`2 \pi`, is taken care of internally where needed). Phase (in radians) will not undergo unit conversion.

    For the driving field, we specify a detuning of
    :math:`\delta = 1 \times 2 \pi \text{MHz}`, and an
    amplitude :math:`\Omega(t)` defined by a sinusoidal oscillation, squared to ensure a positve amplitude
    (a requirement for some hardware implementations). The maximum amplitude will dependent on the parameter ``p``
    passed to the amplitude function later, and should also be passed in units of MHz. We introduce a small phase
    shift as well, on the order of 1 rad.

    .. code-block:: python

        atom_coordinates = [[0, 0], [0, 4], [4, 0], [4, 4]]
        wires = [0, 1, 2, 3]
        H_i = qml.pulse.rydberg_interaction(atom_coordinates, wires)

        amplitude = lambda p, t: p * jnp.sin(jnp.pi * t) ** 2
        phase = 0.25
        detuning = 1.
        H_d = qml.pulse.rydberg_drive(amplitude, phase, detuning, wires)

    >>> H_i
    ParametrizedHamiltonian: terms=6
    >>> H_d
    ParametrizedHamiltonian: terms=3

    The first two terms of the drive Hamiltonian ``H_d`` correspond to the first sum (the sine and cosine terms),
    describing drive between the ground and excited states. The third term corresponds to the shift term
    due to detuning from resonance. This drive term corresponds to a global drive that acts on all 4 wires of
    the device.

    The full Hamiltonian evolution and expectation value measurement can be executed in a ``QNode``:

    .. code-block:: python

        dev = qml.device("default.qubit.jax", wires=wires)
        @qml.qnode(dev, interface="jax")
        def circuit(params):
            qml.evolve(H_i + H_d)(params, t=[0, 0.5])
            return qml.expval(qml.PauliZ(0))

    Here we set a maximum amplitude of :math:`2.4 \times 2 \pi \text{MHz}`, and calculate the result of running the pulse program:

    >>> params = [2.4]
    >>> circuit(params)
    Array(0.78301974, dtype=float64)
    >>> jax.grad(circuit)(params)
    [Array(-0.6250622, dtype=float64)]

    We can also create a Hamiltonian with local drives. The following circuit corresponds to the
    evolution where additional local drives acting on wires ``0`` and ``1`` respectively are added to the
    Hamiltonian:

    .. code-block:: python

        amplitude_local_0 = lambda p, t: p[0] * jnp.sin(2 * jnp.pi * t) ** 2 + p[1]
        phase_local_0 = jnp.pi / 4
        detuning_local_0 = lambda p, t: p * jnp.exp(-0.25 * t)
        H_local_0 = qml.pulse.rydberg_drive(amplitude_local_0, phase_local_0, detuning_local_0, [0])

        amplitude_local_1 = lambda p, t: jnp.cos(jnp.pi * t) ** 2 + p
        phase_local_1 = jnp.pi
        detuning_local_1 = lambda p, t: jnp.sin(jnp.pi * t) + p
        H_local_1 = qml.pulse.rydberg_drive(amplitude_local_1, phase_local_1, detuning_local_1, [1])

        H = H_i + H_d + H_local_0 + H_local_1

        @jax.jit
        @qml.qnode(dev, interface="jax")
        def circuit_local(params):
            qml.evolve(H)(params, t=[0, 0.5])
            return qml.expval(qml.PauliZ(0))

        p_global = 2.4
        p_local_amp_0 = [1.3, -2.0]
        p_local_det_0 = -1.5
        p_local_amp_1 = -0.9
        p_local_det_1 = 3.1
        params = [p_global, p_local_amp_0, p_local_det_0, p_local_amp_1, p_local_det_1]


    >>> circuit_local(params)
    Array(0.62640288, dtype=float64)
    >>> jax.grad(circuit_local)(params)
    [Array(1.07614151, dtype=float64),
     [Array(0.36370049, dtype=float64, weak_type=True),
      Array(0.91057498, dtype=float64, weak_type=True)],
     Array(1.3166343, dtype=float64),
     Array(-0.11102892, dtype=float64),
     Array(-0.02205843, dtype=float64)]
    """
    wires = Wires(wires)
    trivial_detuning = not callable(detuning) and qml.math.isclose(detuning, 0.0)

    if not callable(amplitude) and qml.math.isclose(amplitude, 0.0):
        if trivial_detuning:
            raise ValueError(
                "Expected non-zero value for at least one of either amplitude or detuning, but "
                f"received amplitude={amplitude} and detuning={detuning}. All terms are zero."
            )

        amplitude_term = HardwareHamiltonian([], [])

    else:
        amplitude_term = drive(amplitude, phase, wires)

    detuning_obs, detuning_coeffs = [], []
    if not trivial_detuning:
        # 2pi factors are to convert detuning frequency to angular frequency
        detuning_obs.append(
            # Global phase from the number operator
            -0.5 * sum(qml.Identity(wire) for wire in wires) * np.pi * 2
            # Equivalent of the number operator up to the global phase above
            + 0.5 * sum(qml.PauliZ(wire) for wire in wires) * np.pi * 2
        )
        detuning_coeffs.append(detuning)

    detuning_term = HardwareHamiltonian(detuning_coeffs, detuning_obs)
    pulses = [HardwarePulse(amplitude, phase, detuning, wires)]

    drive_term = amplitude_term + detuning_term
    drive_term.pulses = pulses

    return drive_term


@dataclass
class RydbergSettings:
    """Dataclass that contains the information of a Rydberg setup.

    Args:
        register (list): coordinates of atoms
        interaction_coeff (float): interaction coefficient C6 from C6/(Rij)**6 term in :func:`rydberg_interaction`
    """

    register: list
    interaction_coeff: float = 0.0

    def __eq__(self, other):
        return (
            qml.math.all(self.register == other.register)
            and self.interaction_coeff == other.interaction_coeff
        )

    def __add__(self, other):
        if other is not None:
            raise ValueError(
                "Cannot add two `HardwareHamiltonian` instances with an interaction term. "
                f"Obtained two instances with settings {self} and {other}. "
                "You most likely tried to add two terms generated by `qml.pulse.rydberg_interaction`"
            )

        return self

    __radd__ = __add__
